#!/usr/bin/env python3
"""Upgrades CoLRev projects."""
from __future__ import annotations

import json
import re
import shutil
import typing
from importlib.metadata import version
from pathlib import Path

import git
import pandas as pd
from tqdm import tqdm
from yaml import safe_load

import colrev.env.utils
import colrev.exceptions as colrev_exceptions
import colrev.process.operation
from colrev.constants import Colors
from colrev.constants import Fields
from colrev.constants import FieldValues
from colrev.constants import Filepaths
from colrev.constants import OperationsType
from colrev.constants import RecordState
from colrev.writer.write_utils import to_string

# pylint: disable=too-few-public-methods
# pylint: disable=line-too-long


class Upgrade(colrev.process.operation.Operation):
    """Upgrade a CoLRev project"""

    repo: git.Repo

    type = OperationsType.check

    def __init__(
        self,
        *,
        review_manager: colrev.review_manager.ReviewManager,
    ) -> None:
        prev_force_mode = review_manager.force_mode
        review_manager.force_mode = True
        super().__init__(
            review_manager=review_manager,
            operations_type=self.type,
            notify_state_transition_operation=False,
        )
        review_manager.force_mode = prev_force_mode
        self.review_manager = review_manager

    def _move_file(self, source: Path, target: Path) -> None:
        target.parent.mkdir(exist_ok=True, parents=True)
        if source.is_file():
            shutil.move(str(source), str(self.review_manager.path / target))
            self.repo.index.remove([str(source)])
            self.repo.index.add([str(target)])

    def _load_settings_dict(self) -> dict:
        settings_path = self.review_manager.paths.settings
        if not settings_path.is_file():
            raise colrev_exceptions.CoLRevException()
        with open(settings_path, encoding="utf-8") as file:
            settings = json.load(file)
        return settings

    def _save_settings(self, settings: dict) -> None:
        with open("settings.json", "w", encoding="utf-8") as outfile:
            json.dump(settings, outfile, indent=4)
        self.repo.index.add(["settings.json"])

    def load_records_dict(self) -> dict:
        """
        Load the records dictionary from a file and parse it using the bibtex parser.

        Returns:
            dict: The loaded records dictionary.
        """
        records = colrev.loader.load_utils.load(
            filename=Path("data/records.bib"),
            logger=self.review_manager.logger,
        )

        return records

    def save_records_dict(self, records: dict) -> None:
        """
        Save the records dictionary to a file and add it to the repository index.

        Args:
            records (dict): The records dictionary to save.
        """
        bibtex_str = to_string(records_dict=records, implementation="bib")
        with open("data/records.bib", "w", encoding="utf-8") as out:
            out.write(bibtex_str + "\n")
        self.repo.index.add(["data/records.bib"])

    def main(self) -> None:
        """Upgrade a CoLRev project (main entrypoint)"""

        try:
            self.repo = git.Repo(str(self.review_manager.path))
            self.repo.iter_commits()
        except ValueError:
            # Git repository has no initial commit
            return

        settings = self._load_settings_dict()
        settings_version_str = settings["project"]["colrev_version"]

        settings_version = CoLRevVersion(settings_version_str)
        # Start with the first step if the version is older:
        settings_version = max(settings_version, CoLRevVersion("0.7.0"))
        installed_colrev_version = CoLRevVersion(version("colrev"))

        # version: indicates from which version on the migration should be applied
        migration_scripts: typing.List[typing.Dict[str, typing.Any]] = [
            {
                "version": CoLRevVersion("0.7.0"),
                "target_version": CoLRevVersion("0.7.1"),
                "script": self._migrate_0_7_0,
                "released": True,
            },
            {
                "version": CoLRevVersion("0.7.1"),
                "target_version": CoLRevVersion("0.8.0"),
                "script": self._migrate_0_7_1,
                "released": True,
            },
            # Note : we may add a flag to update to pre-released versions
            {
                "version": CoLRevVersion("0.8.0"),
                "target_version": CoLRevVersion("0.8.1"),
                "script": self._migrate_0_8_0,
                "released": True,
            },
            {
                "version": CoLRevVersion("0.8.1"),
                "target_version": CoLRevVersion("0.8.2"),
                "script": self._migrate_0_8_1,
                "released": True,
            },
            {
                "version": CoLRevVersion("0.8.2"),
                "target_version": CoLRevVersion("0.8.3"),
                "script": self._migrate_0_8_2,
                "released": True,
            },
            {
                "version": CoLRevVersion("0.8.3"),
                "target_version": CoLRevVersion("0.8.4"),
                "script": self._migrate_0_8_3,
                "released": True,
            },
            {
                "version": CoLRevVersion("0.8.4"),
                "target_version": CoLRevVersion("0.9.0"),
                "script": self._migrate_0_8_4,
                "released": True,
            },
            {
                "version": CoLRevVersion("0.9.0"),
                "target_version": CoLRevVersion("0.9.1"),
                "script": self._migrate_0_9_1,
                "released": True,
            },
            {
                "version": CoLRevVersion("0.9.2"),
                "target_version": CoLRevVersion("0.9.3"),
                "script": self._migrate_0_9_3,
                "released": True,
            },
            {
                "version": CoLRevVersion("0.10.0"),
                "target_version": CoLRevVersion("0.10.1"),
                "script": self._migrate_0_10_1,
                "released": True,
            },
            {
                "version": CoLRevVersion("0.10.1"),
                "target_version": CoLRevVersion("0.10.2"),
                "script": self._migrate_0_10_2,
                "released": True,
            },
            {
                "version": CoLRevVersion("0.10.2"),
                "target_version": CoLRevVersion("0.11.0"),
                "script": self._migrate_0_11_0,
                "released": True,
            },
            {
                "version": CoLRevVersion("0.11.0"),
                "target_version": CoLRevVersion("0.12.0"),
                "script": self._migrate_0_12_0,
                "released": True,
            },
            {
                "version": CoLRevVersion("0.12.0"),
                "target_version": CoLRevVersion("0.13.0"),
                "script": self._migrate_0_13_0,
                "released": True,
            },
        ]
        self.review_manager.logger.info(
            "Colrev version installed:           %s", installed_colrev_version
        )
        self.review_manager.logger.info(
            "Colrev version in project settings: %s", settings_version
        )
        # Note: we should always update the colrev_version in settings.json because the
        # checker._check_software requires the settings version and
        # the installed version to be identical

        # skipping_versions_before_settings_version = True
        run_migration = False
        while migration_scripts:
            migrator = migration_scripts.pop(0)
            # Activate run_migration for the current settings_version
            if (
                migrator["target_version"] >= settings_version
            ):  # settings_version == migrator["version"] or
                run_migration = True
            if not run_migration:
                continue
            if installed_colrev_version == settings_version and migrator["released"]:
                break

            migration_script = migrator["script"]
            self.review_manager.logger.info(
                "Upgrade to: %s", migrator["target_version"]
            )
            if migrator["released"]:
                self._print_release_notes(selected_version=migrator["target_version"])

            updated = migration_script()
            if not updated:
                continue

        if not run_migration:
            print("migration not run")
            return

        settings = self._load_settings_dict()
        settings["project"]["colrev_version"] = str(installed_colrev_version)
        self._save_settings(settings)

        if self.repo.is_dirty():
            msg = f"Upgrade to CoLRev {installed_colrev_version}"
            if not migrator["released"]:
                msg += " (pre-release)"
            review_manager = colrev.review_manager.ReviewManager()
            review_manager.dataset.create_commit(
                msg=msg,
            )

    def _print_release_notes(self, *, selected_version: CoLRevVersion) -> None:
        filedata = colrev.env.utils.get_package_file_content(
            module="colrev", filename=Path("../CHANGELOG.md")
        )
        active, printed = False, False
        if filedata:
            for line in filedata.decode("utf-8").split("\n"):
                if str(selected_version) in line:
                    active = True
                    print(f"{Colors.ORANGE}Release notes v{selected_version}")
                    continue
                if line.startswith("## "):
                    active = False
                if active:
                    print(line)
                    printed = True
        if not printed:
            print(f"{Colors.ORANGE}No release notes")
        print(f"{Colors.END}")

    def _migrate_0_7_0(self) -> bool:
        pre_commit_contents = Path(".pre-commit-config.yaml").read_text(
            encoding="utf-8"
        )
        if "ci:" not in pre_commit_contents:
            pre_commit_contents = pre_commit_contents.replace(
                "repos:",
                "ci:\n    skip: [colrev-hooks-format, colrev-hooks-check]\n\nrepos:",
            )
            with open(".pre-commit-config.yaml", "w", encoding="utf-8") as file:
                file.write(pre_commit_contents)
        self.repo.index.add([".pre-commit-config.yaml"])
        return self.repo.is_dirty()

    def _migrate_0_7_1(self) -> bool:
        settings_content = (self.review_manager.path / Path("settings.json")).read_text(
            encoding="utf-8"
        )
        settings_content = settings_content.replace("colrev_built_in.", "colrev.")

        with open(Path("settings.json"), "w", encoding="utf-8") as file:
            file.write(settings_content)

        self.repo.index.add(["settings.json"])
        self.review_manager.load_settings()
        if self.review_manager.settings.is_curated_masterdata_repo():
            self.review_manager.settings.project.delay_automated_processing = False
        self.review_manager.save_settings()

        self._move_file(source=Path("data/paper.md"), target=Path("data/data/paper.md"))
        self._move_file(
            source=Path("data/APA-7.docx"), target=Path("data/data/APA-7.docx")
        )
        self._move_file(
            source=Path("data/non_sample_references.bib"),
            target=Path("data/data/non_sample_references.bib"),
        )

        return self.repo.is_dirty()

    def _migrate_0_8_0(self) -> bool:
        Path(".github/workflows/").mkdir(exist_ok=True, parents=True)

        if "colrev/curated_metadata" in str(self.review_manager.path):
            Path(".github/workflows/colrev_update.yml").unlink(missing_ok=True)
            colrev.env.utils.retrieve_package_file(
                template_file=Path("ops/init/colrev_update_curation.yml"),
                target=Path(".github/workflows/colrev_update.yml"),
            )
            self.repo.index.add([".github/workflows/colrev_update.yml"])
        else:
            Path(".github/workflows/colrev_update.yml").unlink(missing_ok=True)
            colrev.env.utils.retrieve_package_file(
                template_file=Path("ops/init/colrev_update.yml"),
                target=Path(".github/workflows/colrev_update.yml"),
            )
            self.repo.index.add([".github/workflows/colrev_update.yml"])

        Path(".github/workflows/pre-commit.yml").unlink(missing_ok=True)
        colrev.env.utils.retrieve_package_file(
            template_file=Path("ops/init/pre-commit.yml"),
            target=Path(".github/workflows/pre-commit.yml"),
        )
        self.repo.index.add([".github/workflows/pre-commit.yml"])
        return self.repo.is_dirty()

    def _migrate_0_8_1(self) -> bool:
        Path(".github/workflows/").mkdir(exist_ok=True, parents=True)
        if "colrev/curated_metadata" in str(self.review_manager.path):
            Path(".github/workflows/colrev_update.yml").unlink(missing_ok=True)
            colrev.env.utils.retrieve_package_file(
                template_file=Path("ops/init/colrev_update_curation.yml"),
                target=Path(".github/workflows/colrev_update.yml"),
            )
            self.repo.index.add([".github/workflows/colrev_update.yml"])
        else:
            Path(".github/workflows/colrev_update.yml").unlink(missing_ok=True)
            colrev.env.utils.retrieve_package_file(
                template_file=Path("ops/init/colrev_update.yml"),
                target=Path(".github/workflows/colrev_update.yml"),
            )
            self.repo.index.add([".github/workflows/colrev_update.yml"])

        settings = self._load_settings_dict()
        settings["project"]["auto_upgrade"] = True
        self._save_settings(settings)

        return self.repo.is_dirty()

    def _migrate_0_8_2(self) -> bool:
        records = self.review_manager.dataset.load_records_dict()

        for record_dict in tqdm(records.values()):
            if "colrev_pdf_id" not in record_dict:
                continue
            if not record_dict["colrev_pdf_id"].startswith("cpid1:"):
                continue
            if not Path(record_dict.get("file", "")).is_file():
                continue

            pdf_path = Path(record_dict["file"])
            colrev_pdf_id = colrev.record.record.Record.get_colrev_pdf_id(pdf_path)
            # pylint: disable=colrev-missed-constant-usage
            record_dict["colrev_pdf_id"] = colrev_pdf_id

        self.review_manager.dataset.save_records_dict(records)

        return self.repo.is_dirty()

    def _migrate_0_8_3(self) -> bool:
        # pylint: disable=too-many-branches
        settings = self._load_settings_dict()
        settings["prep"]["defects_to_ignore"] = []
        if "curated_metadata" in str(self.review_manager.path):
            settings["prep"]["defects_to_ignore"] = [
                "record-not-in-toc",
                "inconsistent-with-url-metadata",
            ]
        else:
            settings["prep"]["defects_to_ignore"] = ["inconsistent-with-url-metadata"]

        for p_round in settings["prep"]["prep_rounds"]:
            p_round["prep_package_endpoints"] = [
                x
                for x in p_round["prep_package_endpoints"]
                if x["endpoint"] != "colrev.global_ids_consistency_check"
            ]
        self._save_settings(settings)
        self.review_manager = colrev.review_manager.ReviewManager(
            path_str=str(self.review_manager.path), force_mode=True
        )
        self.review_manager.load_settings()
        self.review_manager.get_load_operation()
        records = self.review_manager.dataset.load_records_dict()
        quality_model = self.review_manager.get_qm()

        # delete the masterdata provenance notes and apply the new quality model
        # replace not_missing > not-missing
        for record_dict in tqdm(records.values()):
            if Fields.MD_PROV not in record_dict:
                continue
            not_missing_fields = []
            for key, prov in record_dict[Fields.MD_PROV].items():
                if "not-missing" in prov["note"]:
                    not_missing_fields.append(key)
                prov["note"] = ""
            for key in not_missing_fields:
                record_dict[Fields.MD_PROV][key]["note"] = "not-missing"
            if "cited_by_file" in record_dict:
                del record_dict["cited_by_file"]
            if "cited_by_id" in record_dict:
                del record_dict["cited_by_id"]
            if "tei_id" in record_dict:
                del record_dict["tei_id"]
            if Fields.D_PROV in record_dict:
                if "cited_by_file" in record_dict[Fields.D_PROV]:
                    del record_dict[Fields.D_PROV]["cited_by_file"]
                if "cited_by_id" in record_dict[Fields.D_PROV]:
                    del record_dict[Fields.D_PROV]["cited_by_id"]
                if "tei_id" in record_dict[Fields.D_PROV]:
                    del record_dict[Fields.D_PROV]["tei_id"]

            record = colrev.record.record.Record(record_dict)
            prior_state = record.data[Fields.STATUS]
            record.run_quality_model(quality_model)
            if prior_state == RecordState.rev_prescreen_excluded:
                record.data[  # pylint: disable=colrev-direct-status-assign
                    Fields.STATUS
                ] = RecordState.rev_prescreen_excluded
        self.review_manager.dataset.save_records_dict(records)
        return self.repo.is_dirty()

    def _migrate_0_8_4(self) -> bool:
        records = self.review_manager.dataset.load_records_dict()
        for record in records.values():
            if Fields.EDITOR not in record.get(Fields.D_PROV, {}):
                continue
            ed_val = record[Fields.D_PROV][Fields.EDITOR]
            del record[Fields.D_PROV][Fields.EDITOR]
            if FieldValues.CURATED not in record[Fields.MD_PROV]:
                record[Fields.MD_PROV][Fields.EDITOR] = ed_val

        self.review_manager.dataset.save_records_dict(records)

        return self.repo.is_dirty()

    def _migrate_0_9_1(self) -> bool:
        settings = self._load_settings_dict()
        for source in settings["sources"]:
            if "load_conversion_package_endpoint" in source:
                del source["load_conversion_package_endpoint"]
        self._save_settings(settings)
        return self.repo.is_dirty()

    # pylint: disable=too-many-branches
    def _migrate_0_9_3(self) -> bool:
        settings = self._load_settings_dict()
        for source in settings["sources"]:
            if source["endpoint"] == "colrev.crossref":
                if Fields.ISSN not in source["search_parameters"].get("scope", {}):
                    continue
                if isinstance(source["search_parameters"]["scope"][Fields.ISSN], str):
                    source["search_parameters"]["scope"][Fields.ISSN] = [
                        source["search_parameters"]["scope"][Fields.ISSN]
                    ]

        self._save_settings(settings)

        records = self.load_records_dict()
        for record_dict in records.values():
            if "pubmedid" in record_dict:
                record = colrev.record.record.Record(record_dict)
                record.rename_field(key="pubmedid", new_key="colrev.pubmed.pubmedid")

            if "pii" in record_dict:
                record = colrev.record.record.Record(record_dict)
                record.rename_field(key="pii", new_key="colrev.pubmed.pii")

            if "pmc" in record_dict:
                record = colrev.record.record.Record(record_dict)
                record.rename_field(key="pmc", new_key="colrev.pubmed.pmc")

            if "label_included" in record_dict:
                record = colrev.record.record.Record(record_dict)
                record.rename_field(
                    key="label_included",
                    new_key="colrev.synergy_datasets.label_included",
                )
            if "method" in record_dict:
                record = colrev.record.record.Record(record_dict)
                record.rename_field(
                    key="method", new_key="colrev.synergy_datasets.method"
                )

            if "dblp_key" in record_dict:
                record = colrev.record.record.Record(record_dict)
                record.rename_field(key="dblp_key", new_key=Fields.DBLP_KEY)
            if "wos_accession_number" in record_dict:
                record = colrev.record.record.Record(record_dict)
                record.rename_field(
                    key="wos_accession_number",
                    new_key=Fields.WEB_OF_SCIENCE_ID,
                )
            if "sem_scholar_id" in record_dict:
                record = colrev.record.record.Record(record_dict)
                record.rename_field(
                    key="sem_scholar_id", new_key=Fields.SEMANTIC_SCHOLAR_ID
                )

            if "openalex_id" in record_dict:
                record = colrev.record.record.Record(record_dict)
                record.rename_field(key="openalex_id", new_key="colrev.open_alex.id")

        self.save_records_dict(records)

        return self.repo.is_dirty()

    # pylint: disable=too-many-branches
    def _migrate_0_10_1(self) -> bool:
        prep_replacements = {
            "colrev.open_alex_prep": "colrev.open_alex",
            "colrev.get_masterdata_from_dblp": "colrev.dblp",
            "colrev.crossref_metadata_prep": "colrev.crossref_metadata_prep",
            "colrev.get_masterdata_from_crossref": "colrev.crossref",
            "colrev.get_masterdata_from_europe_pmc": "colrev.europe_pmc",
            "colrev.get_masterdata_from_pubmed": "colrev.pubmed",
            "colrev.get_masterdata_from_open_library": "colrev.open_library",
            "colrev.curation_prep": "colrev.colrev_curation",
            "colrev.get_masterdata_from_local_index": "colrev.local_index",
        }

        settings = self._load_settings_dict()
        for prep_round in settings["prep"]["prep_rounds"]:
            for prep_package in prep_round["prep_package_endpoints"]:
                for old, new in prep_replacements.items():
                    if prep_package["endpoint"] == old:
                        prep_package["endpoint"] = new
        for source in settings["sources"]:
            if source["endpoint"] == "colrev.pdfs_dir":
                source["endpoint"] = "colrev.files_dir"
            if (
                source["endpoint"] == "colrev.dblp"
                and "scope" in source["search_parameters"]
            ):
                if "query" in source["search_parameters"]:
                    source["search_type"] = "API"
                else:
                    source["search_type"] = "TOC"

            if (
                source["endpoint"] == "colrev.crossref"
                and "scope" in source["search_parameters"]
            ):
                if "query" in source["search_parameters"]:
                    source["search_type"] = "API"
                else:
                    source["search_type"] = "TOC"

            if "data/search/md_" in source["filename"]:
                source["search_type"] = "MD"
            if source["search_type"] == "PDFS":
                source["search_type"] = "FILES"

        self._save_settings(settings)
        return self.repo.is_dirty()

    def _migrate_0_10_2(self) -> bool:
        paper_md_path = Path("data/data/paper.md")
        if paper_md_path.is_file():
            paper_md_content = paper_md_path.read_text(encoding="utf-8")
            paper_md_content = paper_md_content.replace(
                "data/records.bib", "data/data/sample_references.bib"
            )
            paper_md_path.write_text(paper_md_content, encoding="utf-8")
            self.repo.index.add([str(paper_md_path)])

        return self.repo.is_dirty()

    def _migrate_0_11_0(self) -> bool:
        settings = self._load_settings_dict()
        if settings["project"]["review_type"] == "curated_masterdata":
            settings["project"]["review_type"] = "colrev.curated_masterdata"

        if "dedupe" in settings:
            settings["dedupe"].pop("same_source_merges", None)

        settings["pdf_get"]["defects_to_ignore"] = []

        settings["pdf_prep"]["pdf_prep_package_endpoints"] = [
            {"endpoint": "colrev.ocrmypdf"},
            {"endpoint": "colrev.grobid_tei"},
        ] + [
            x
            for x in settings["pdf_prep"]["pdf_prep_package_endpoints"]
            if x["endpoint"]
            not in [
                "colrev.check_ocr",
                "colrev.pdf_check_ocr",
                "colrev.validate_pdf_metadata",
                "colrev.validate_completeness",
                "colrev.create_tei",
                "colrev.tei_prep",
            ]
        ]

        if settings["project"]["review_type"] == "curated_masterdata":
            Path(".github/workflows/colrev_update.yml").unlink(missing_ok=True)
            colrev.env.utils.retrieve_package_file(
                template_file=Path(
                    "packages/review_type/curated_masterdata/curations_github_colrev_update.yml"
                ),
                target=Path(".github/workflows/colrev_update.yml"),
            )
            self.repo.index.add([".github/workflows/colrev_update.yml"])
        else:
            Path(".github/workflows/colrev_update.yml").unlink(missing_ok=True)
            colrev.env.utils.retrieve_package_file(
                template_file=Path("ops/init/colrev_update.yml"),
                target=Path(".github/workflows/colrev_update.yml"),
            )
            self.repo.index.add([".github/workflows/colrev_update.yml"])

        for p_round in settings["prep"]["prep_rounds"]:
            p_round["prep_package_endpoints"] = [
                x
                for x in p_round["prep_package_endpoints"]
                if x["endpoint"] != "colrev.resolve_crossrefs"
            ]
            if settings["project"]["review_type"] == "colrev.curated_masterdata":
                p_round["prep_package_endpoints"] = [
                    {"endpoint": "colrev.colrev_curation"}
                ] + p_round["prep_package_endpoints"]

        self._save_settings(settings)

        records = self.load_records_dict()
        for record_dict in records.values():
            if Fields.MD_PROV in record_dict:
                for key, value in record_dict[Fields.MD_PROV].items():
                    record_dict[Fields.MD_PROV][key]["note"] = value["note"].replace(
                        "not-missing", "IGNORE:missing"
                    )
        self.save_records_dict(records)

        return self.repo.is_dirty()

    def _migrate_0_12_0(self) -> bool:
        registry_yaml = Filepaths.LOCAL_ENVIRONMENT_DIR.joinpath(Path("registry.yaml"))

        def _cast_values_to_str(data) -> dict:  # type: ignore
            result = {}
            for key, value in data.items():
                if isinstance(value, dict):
                    result[key] = _cast_values_to_str(value)
                elif isinstance(value, list):
                    result[key] = [_cast_values_to_str(v) for v in value]  # type: ignore
                else:
                    result[key] = str(value)  # type: ignore
            return result

        if registry_yaml.is_file():
            backup_file = Path(str(registry_yaml) + ".bk")
            print(
                f"Found a yaml file, converting to json, it will be backed up as {backup_file}"
            )
            with open(registry_yaml, encoding="utf8") as file:
                environment_registry_df = pd.json_normalize(safe_load(file))
                repos = environment_registry_df.to_dict("records")
                environment_registry = {
                    "local_index": {
                        "repos": repos,
                    },
                    "packages": {},
                }
                Filepaths.REGISTRY_FILE.parents[0].mkdir(parents=True, exist_ok=True)
                with open(Filepaths.REGISTRY_FILE, "w", encoding="utf8") as file:
                    json.dump(
                        dict(_cast_values_to_str(environment_registry)),
                        indent=4,
                        fp=file,
                    )
                shutil.move(str(registry_yaml), str(backup_file))

        return self.repo.is_dirty()

    def _migrate_0_13_0(self) -> bool:
        # Rename "warning" to "colrev.dblp.warning" in all DBLP search_sources

        settings = self._load_settings_dict()
        for source in settings["sources"]:
            if source["endpoint"] == "colrev.dblp":
                records = colrev.loader.load_utils.load(
                    filename=Path(source["filename"]),
                    logger=self.review_manager.logger,
                )
                for record_dict in records.values():
                    if "warning" in record_dict:
                        record_dict["colrev.dblp.warning"] = record_dict.pop("warning")
                    record_dict.pop("metadata_source", None)

                bibtex_str = to_string(records_dict=records, implementation="bib")
                with open(source["filename"], "w", encoding="utf-8") as out:
                    out.write(bibtex_str + "\n")
                self.repo.index.add([source["filename"]])

        # Add "colrev.ref_check" to data endpoints
        if "colrev.ref_check" not in [
            e["endpoint"] for e in settings["data"]["data_package_endpoints"]
        ]:
            settings["data"]["data_package_endpoints"].append(
                {"endpoint": "colrev.ref_check"}
            )

        # Remove "inconsistent-with-url-metadata" from settings["prep"]["defects_to_ignore"]
        settings["prep"]["defects_to_ignore"] = [
            d
            for d in settings["prep"]["defects_to_ignore"]
            if d != "inconsistent-with-url-metadata"
        ]
        self._save_settings(settings)

        # Rename LOCAL_ENVIRONMENT_DIR
        if not Filepaths.LOCAL_ENVIRONMENT_DIR.is_dir():
            shutil.move(
                str(Path.home().joinpath("colrev")),
                str(Filepaths.LOCAL_ENVIRONMENT_DIR),
            )

        # add colrev install . for existing .github/workflows/colrev_update.yml
        if Path(".github/workflows/colrev_update.yml").is_file():
            with open(".github/workflows/colrev_update.yml", encoding="utf-8") as file:
                content = file.read()
                if "colrev install" not in content:
                    content = content.replace(
                        "          poetry run --directory ${{ runner.temp }}/colrev colrev search -f",
                        "          poetry run --directory ${{ runner.temp }}/colrev colrev install .\n"
                        + "          poetry run --directory ${{ runner.temp }}/colrev colrev search -f",
                    )

                    with open(
                        ".github/workflows/colrev_update.yml", "w", encoding="utf-8"
                    ) as out:
                        out.write(content)

                    self.repo.index.add([".github/workflows/colrev_update.yml"])

        # replace colrev for .colrev in registry.yaml
        if Filepaths.REGISTRY_FILE.is_file():
            with open(Filepaths.REGISTRY_FILE, encoding="utf-8") as file:
                content = file.read().replace(
                    "/colrev/curated_metadata/", "/.colrev/curated_metadata/"
                )
                with open(Filepaths.REGISTRY_FILE, "w", encoding="utf-8") as out:
                    out.write(content)

        return self.repo.is_dirty()


# Note: we can ask users to make decisions (when defaults are not clear)
# via input() or simply cancel the process (raise a CoLrevException)


class CoLRevVersion:
    """Class for handling the CoLRev version"""

    def __init__(self, version_string: str) -> None:
        if "+" in version_string:
            version_string = version_string[: version_string.find("+")]
        assert re.match(r"\d+\.\d+\.\d+$", version_string)
        self.major = int(version_string[: version_string.find(".")])
        self.minor = int(
            version_string[version_string.find(".") + 1 : version_string.rfind(".")]
        )
        self.patch = int(version_string[version_string.rfind(".") + 1 :])

    def __eq__(self, other) -> bool:  # type: ignore
        return (
            self.major == other.major
            and self.minor == other.minor
            and self.patch == other.patch
        )

    def __lt__(self, other) -> bool:  # type: ignore
        if self.major < other.major:
            return True
        if self.major == other.major and self.minor < other.minor:
            return True
        if (
            self.major == other.major
            and self.minor == other.minor
            and self.patch < other.patch
        ):
            return True
        return False

    def __ge__(self, other) -> bool:  # type: ignore
        if self.major > other.major:
            return True
        if self.major == other.major and self.minor > other.minor:
            return True
        if (
            self.major == other.major
            and self.minor == other.minor
            and self.patch > other.patch
        ):
            return True
        return False

    def __str__(self) -> str:
        return f"{self.major}.{self.minor}.{self.patch}"



#! /usr/bin/env python
"""CoLRev push operation: Push project and record corrections."""
from __future__ import annotations

import json
from pathlib import Path

import colrev.process.operation
from colrev.constants import Colors
from colrev.constants import EndpointType
from colrev.constants import Fields
from colrev.constants import OperationsType


class Push(colrev.process.operation.Operation):
    """Push the project and record corrections"""

    type = OperationsType.check

    def __init__(self, *, review_manager: colrev.review_manager.ReviewManager) -> None:
        super().__init__(
            review_manager=review_manager,
            operations_type=self.type,
        )

    @colrev.process.operation.Operation.decorate()
    def main(
        self,
        *,
        records_only: bool = False,
        project_only: bool = False,
        all_records: bool = False,
    ) -> None:
        """Push a CoLRev project and records (main entrypoint)"""

        if project_only:
            self._push_project()

        if records_only:
            self._push_record_corrections(all_records)

    def _push_project(self) -> None:
        git_repo = self.review_manager.dataset.get_repo()
        origin = git_repo.remotes.origin
        self.review_manager.logger.info(f"Push changes to {git_repo.remotes.origin}")
        origin.push()

    def _get_change_sets(self) -> dict:
        self.review_manager.logger.info("Collect corrections")

        search_source_mappings = {
            x.get_origin_prefix(): x.endpoint
            for x in self.review_manager.settings.sources
        }

        # group by target-repo to bundle changes in a commit
        change_sets = {}  # type: ignore
        corrections_path = self.review_manager.paths.corrections
        for correction_path in corrections_path.glob("*.json"):
            with open(correction_path, encoding="utf8") as json_file:
                output = json.load(json_file)

            record = output["original_record"]
            for source_origin in record[Fields.ORIGIN]:
                # note : simple heuristic / should be based on the SearchSources
                # (objects - whether they offer correction functionality)
                source_prefix = source_origin[: source_origin.find("/")]
                search_source = search_source_mappings[
                    source_origin[: source_origin.find("/")]
                ]

                if search_source in ["colrev.unknown_source"]:
                    continue

                output[Fields.FILE] = correction_path
                if source_prefix not in change_sets:
                    change_sets[source_prefix] = []

                change_sets[source_prefix].append(output)

        return change_sets

    def _push_record_corrections(self, all_records: bool) -> None:
        """Push corrections of records"""

        change_sets = self._get_change_sets()
        package_manager = self.review_manager.get_package_manager()

        for source_prefix, change_itemsets in change_sets.items():
            source_l = [
                s
                for s in self.review_manager.settings.sources
                if source_prefix == s.get_origin_prefix()
            ]
            if not source_l:
                continue

            source = source_l[0]

            search_source_class = package_manager.get_package_endpoint_class(
                package_type=EndpointType.search_source,
                package_identifier=source.endpoint,
            )
            endpoint = search_source_class(
                source_operation=self, settings=source.get_dict()
            )

            correct_function = getattr(endpoint, "apply_correction", None)
            if callable(correct_function):
                self.review_manager.logger.info(
                    f"{Colors.GREEN}Push record corrections to {source_prefix}{Colors.END}"
                )
                endpoint.apply_correction(change_itemsets=change_itemsets)  # type: ignore

            elif all_records:
                # Else: use the share_corrections functionality:
                self.review_manager.logger.info(
                    f"{Colors.GREEN}Push record corrections to {source_prefix}{Colors.END}"
                )
                self.review_manager.logger.debug(
                    f"No correction function in {source.endpoint}"
                )
                self._share_correction(source=source, change_list=change_itemsets)

    def _share_correction(
        self, *, source: colrev.settings.SearchSource, change_list: list
    ) -> None:
        prepared_change_list = []
        for change in change_list:
            prepared_change_list.append(
                {
                    "record": change["original_record"],
                    "changes": change["changes"],
                }
            )

        corrections = self.review_manager.p_printer.pformat(prepared_change_list)

        text = (
            "Dear Sir or Madam,\n\nwe have noticed potential corrections and "
            + "would like to share them with you.\nThe potentical changes are:\n\n"
            + f"{corrections}\n\nBest regards\n\n"
        )

        if source.endpoint == "colrev.dblp":
            file_path = Path("dblp-corrections-mail.txt")
            dblp_header = (
                "Send to: dblp@dagstuhl.de\n\n"
                + "Subject: Potential correction to DBLP metadata\n\n"
            )

            text = dblp_header + text
            file_path.write_text(text, encoding="utf-8")

            print(f"\nPlease send the e-mail (prepared in the file {file_path})")
            input("Press Enter to confirm")

            for change_item in change_list:
                if Path(change_item[Fields.FILE]).is_file():
                    Path(change_item[Fields.FILE]).unlink()



#!/usr/bin/env python
"""Utility to transform items from semanticscholar into records"""
from __future__ import annotations

import re

from semanticscholar import Paper

import colrev.exceptions as colrev_exceptions
import colrev.record.record
import colrev.record.record_prep
from colrev.constants import ENTRYTYPES
from colrev.constants import Fields

# https://api.semanticscholar.org/api-docs

PUB_TYPE_ENTRYTYPE_MAP = {
    "Review": ENTRYTYPES.ARTICLE,
    "JournalArticle": ENTRYTYPES.ARTICLE,
    "CaseReport": ENTRYTYPES.TECHREPORT,
    "ClinicalTrial": ENTRYTYPES.MISC,
    "Dataset": ENTRYTYPES.MISC,
    "Editorial": ENTRYTYPES.MISC,
    "LettersAndComments": ENTRYTYPES.MISC,
    "MetaAnalysis": ENTRYTYPES.ARTICLE,
    "News": ENTRYTYPES.MISC,
    "Study": ENTRYTYPES.MISC,
    "Book": ENTRYTYPES.BOOK,
    "BookSection": ENTRYTYPES.INBOOK,
    "Conference": ENTRYTYPES.INPROCEEDINGS,
}

SUPPORTED_FIELDS = [
    Fields.ID,
    Fields.SEMANTIC_SCHOLAR_ID,
    Fields.DBLP_KEY,
    Fields.DOI,
    Fields.ENTRYTYPE,
    Fields.ISSN,
    Fields.JOURNAL,
    Fields.BOOKTITLE,
    Fields.VOLUME,
    Fields.PAGES,
    Fields.TITLE,
    Fields.AUTHOR,
    Fields.YEAR,
    Fields.CITED_BY,
    Fields.ABSTRACT,
    Fields.FULLTEXT,
]


def _assign_entrytype(*, record_dict: dict) -> None:
    """Method to assign the ENTRYTYPE"""

    publication_types = record_dict.get("publicationTypes")
    if publication_types is not None and len(publication_types) > 0:
        entrytype = publication_types[0]
    else:
        entrytype = ENTRYTYPES.MISC

    record_dict[Fields.ENTRYTYPE] = PUB_TYPE_ENTRYTYPE_MAP.get(
        entrytype, ENTRYTYPES.MISC
    )
    if record_dict[Fields.ENTRYTYPE] == ENTRYTYPES.MISC and "journal" in record_dict:
        record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.ARTICLE


def _assign_authors(*, record_dict: dict) -> None:
    """Method to assign authors from item"""

    authors = [
        author["name"] for author in record_dict.get("authors", []) if "name" in author
    ]
    record_dict[Fields.AUTHOR] = (
        colrev.record.record_prep.PrepRecord.format_author_field(" and ".join(authors))
    )


def _assign_issn(*, record_dict: dict) -> None:
    """Method to assign the issn"""

    issn = (record_dict.get("publicationVenue", {}) or {}).get("issn")
    if issn:
        record_dict[Fields.ISSN] = issn


def _assign_ids(*, record_dict: dict) -> None:
    """Method to assign external IDs from item"""
    ext_ids = record_dict.get("externalIds", {})

    if "DOI" in ext_ids:
        record_dict[Fields.DOI] = ext_ids["DOI"].upper()
    if "DBLP" in ext_ids:
        record_dict[Fields.DBLP_KEY] = str(ext_ids["DBLP"])


def _assign_abstract(*, record_dict: dict) -> None:
    """Method to assign external IDs from item"""
    abstract = record_dict.get("abstract", "")
    if abstract is not None:
        record_dict[Fields.ABSTRACT] = abstract.replace("\n", " ").lstrip().rstrip()
    else:
        record_dict[Fields.ABSTRACT] = ""


def _assign_article_fields(*, record_dict: dict) -> None:
    """Method to assign pages and volume from item"""

    if record_dict.get("journal") is None:
        return

    if "volume" in record_dict["journal"]:
        record_dict[Fields.VOLUME] = record_dict["journal"]["volume"]
    if "pages" in record_dict["journal"]:
        pages = record_dict["journal"]["pages"]
        if "-" in pages:
            try:
                pages = re.sub(r"\s+|[a-zA-Z]+", "", pages)
                pages_list = pages.split("-")
                page_a = int(pages_list[0])
                page_b = int(pages_list[1])
                if page_a > page_b:
                    pages = str(page_b) + "--" + str(page_a)
                elif page_a == page_b:
                    pages = str(page_a)
                else:
                    pages = str(page_a) + "--" + str(page_b)
            except ValueError:
                pass

        record_dict[Fields.PAGES] = pages

    record_dict[Fields.JOURNAL] = record_dict.get("venue", "")


def _assign_book_fields(*, record_dict: dict) -> None:
    """Method to assign book specific details from item"""
    book_title = record_dict.get("journal", {}).get("name")
    title = record_dict.get(Fields.TITLE, "")

    if book_title and title:
        if (book_title in title) or (title in book_title):
            record_dict[Fields.TITLE] = max(book_title, title, key=len)
        else:
            record_dict[Fields.BOOKTITLE] = book_title


def _assign_inproc_fields(*, record_dict: dict) -> None:
    """Method to assign inproceedings specific details from item"""

    if "journal" in record_dict:
        del record_dict["journal"]

    venue = record_dict.get("venue", "")
    if venue:
        record_dict[Fields.BOOKTITLE] = venue


def _assign_fulltext(*, record_dict: dict) -> None:
    """Method to assign fulltext from item."""
    fulltext_url = (record_dict.get("openAccessPdf") or {}).get("url", "")
    if fulltext_url:
        record_dict[Fields.FULLTEXT] = fulltext_url


def _item_to_record(item: Paper) -> dict:
    """Method to convert the different fields and information within item to record dictionary"""

    record_dict = dict(item)

    record_dict[Fields.SEMANTIC_SCHOLAR_ID] = record_dict["paperId"]
    record_dict[Fields.TITLE] = record_dict.get("title", "")
    _assign_authors(record_dict=record_dict)
    _assign_fulltext(record_dict=record_dict)
    _assign_issn(record_dict=record_dict)
    _assign_ids(record_dict=record_dict)
    _assign_abstract(record_dict=record_dict)
    record_dict[Fields.YEAR] = record_dict.get("year", "")
    record_dict[Fields.CITED_BY] = record_dict.get("citationCount", "")

    _assign_entrytype(record_dict=record_dict)
    if record_dict[Fields.ENTRYTYPE] == ENTRYTYPES.ARTICLE:
        _assign_article_fields(record_dict=record_dict)

    elif record_dict[Fields.ENTRYTYPE] in [ENTRYTYPES.BOOK, ENTRYTYPES.INBOOK]:
        _assign_book_fields(record_dict=record_dict)

    elif record_dict[Fields.ENTRYTYPE] == ENTRYTYPES.INPROCEEDINGS:
        _assign_inproc_fields(record_dict=record_dict)

    return record_dict


def _remove_fields(*, record: dict) -> dict:
    """Method to remove unsupported fields from semanticscholar record"""

    record_dict = {k: v for k, v in record.items() if k in SUPPORTED_FIELDS and v != ""}

    return record_dict


def dict_to_record(*, item: dict) -> colrev.record.record.Record:
    """Convert a semanticscholar item to a record dict"""

    try:
        record_dict = _item_to_record(item=item)
        record_dict = _remove_fields(record=record_dict)
        record = colrev.record.record.Record(record_dict)

    except (IndexError, KeyError) as exc:
        raise colrev_exceptions.RecordNotParsableException(
            f"Exception: Record not parsable: {exc}"
        ) from exc

    return record



#! /usr/bin/env python
"""SearchSource: ACM Digital Library"""
from __future__ import annotations

from pathlib import Path

import zope.interface
from pydantic import Field

import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import Fields
from colrev.constants import SearchSourceHeuristicStatus
from colrev.constants import SearchType

# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.package_manager.interfaces.SearchSourceInterface)
class ACMDigitalLibrarySearchSource:
    """ACM digital Library"""

    settings_class = colrev.package_manager.package_settings.DefaultSourceSettings
    endpoint = "colrev.acm_digital_library"
    # Note : the ID contains the doi
    # "https://dl.acm.org/doi/{{ID}}"
    source_identifier = "doi"
    search_types = [SearchType.DB]

    ci_supported: bool = Field(default=False)
    heuristic_status = SearchSourceHeuristicStatus.supported
    db_url = "https://dl.acm.org/"

    def __init__(
        self, *, source_operation: colrev.process.operation.Operation, settings: dict
    ) -> None:
        self.search_source = self.settings_class(**settings)
        self.review_manager = source_operation.review_manager
        self.operation = source_operation

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for ACM dDigital Library"""

        result = {"confidence": 0.0}
        # Simple heuristic:
        if "publisher = {Association for Computing Machinery}," in data:
            result["confidence"] = 0.7
            return result
        # We may also check whether the ID=doi=url
        return result

    @classmethod
    def add_endpoint(
        cls,
        operation: colrev.ops.search.Search,
        params: str,
    ) -> colrev.settings.SearchSource:
        """Add SearchSource as an endpoint (based on query provided to colrev search --add )"""

        params_dict = {}
        if params:
            for item in params.split(";"):
                key, value = item.split("=")
                params_dict[key] = value

        search_type = operation.select_search_type(
            search_types=cls.search_types, params=params_dict
        )

        if search_type == SearchType.DB:
            search_source = operation.create_db_source(
                search_source_cls=cls,
                params=params_dict,
            )
        else:
            raise NotImplementedError

        operation.add_source_and_search(search_source)
        return search_source

    def search(self, rerun: bool) -> None:
        """Run a search of ACM Digital Library"""

        if self.search_source.search_type == SearchType.DB:
            if self.search_source.filename.suffix in [".bib"]:
                self.operation.run_db_search(  # type: ignore
                    search_source_cls=self.__class__,
                    source=self.search_source,
                )
                return

            raise NotImplementedError
        raise NotImplementedError

    def prep_link_md(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.record.Record:
        """Not implemented"""
        return record

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".bib":

            def field_mapper(record_dict: dict) -> None:
                record_dict.pop("url", None)
                record_dict.pop("publisher", None)
                record_dict.pop("numpages", None)
                record_dict.pop("month", None)

                if "issue_date" in record_dict:
                    record_dict[f"{self.endpoint}.issue_date"] = record_dict.pop(
                        "issue_date"
                    )
                if "location" in record_dict:
                    record_dict[Fields.ADDRESS] = record_dict.pop("location", None)
                if "articleno" in record_dict:
                    record_dict[f"{self.endpoint}.articleno"] = record_dict.pop(
                        "articleno"
                    )

            records = colrev.loader.load_utils.load(
                filename=self.search_source.filename,
                unique_id_field="ID",
                field_mapper=field_mapper,
                logger=self.review_manager.logger,
            )

            return records

        raise NotImplementedError

    # pylint: disable=colrev-missed-constant-usage
    def prepare(
        self, record: colrev.record.record.Record, source: colrev.settings.SearchSource
    ) -> colrev.record.record.Record:
        """Source-specific preparation for ACM Digital Library"""

        return record



#! /usr/bin/env python
"""Checker for author-not-in-pdf."""
from __future__ import annotations

import re
from pathlib import Path

import colrev.env.utils
import colrev.record.qm.quality_model
from colrev.constants import Fields
from colrev.constants import PDFDefectCodes

# pylint: disable=too-few-public-methods

# Note: replaces author_not_in_first_pages


class AuthorNotInPDFChecker:
    """The AuthorNotInPDFChecker"""

    msg = PDFDefectCodes.AUTHOR_NOT_IN_PDF

    def __init__(
        self, quality_model: colrev.record.qm.quality_model.QualityModel
    ) -> None:
        self.quality_model = quality_model

    def run(self, *, record: colrev.record.record.Record) -> None:
        """Run the author-not-in-pdf checks"""

        if (
            Fields.FILE not in record.data
            or Path(record.data[Fields.FILE]).suffix != ".pdf"
            or Fields.AUTHOR not in record.data
            or record.data[Fields.TEXT_FROM_PDF] == ""
            or record.ignored_defect(key=Fields.FILE, defect=self.msg)
        ):
            return

        if not self._author_in_pdf(record=record):
            record.add_field_provenance_note(key=Fields.FILE, note=self.msg)
        else:
            record.remove_field_provenance_note(key=Fields.FILE, note=self.msg)

    def _author_in_pdf(self, *, record: colrev.record.record.Record) -> bool:
        # Many editorials do not have authors in the PDF (or on the last page)
        if "editorial" in record.data.get(Fields.TITLE, "").lower():
            return True

        text = record.data[Fields.TEXT_FROM_PDF].lower()
        text = colrev.env.utils.remove_accents(text)
        text = re.sub("[^a-zA-Z ]+", "", text)
        text = text.replace("ue", "u").replace("ae", "a").replace("oe", "o")

        authors_str = record.data.get(Fields.AUTHOR, "").lower()
        authors_str = (
            authors_str.replace("ue", "u").replace("ae", "a").replace("oe", "o")
        )
        authors_str = colrev.env.utils.remove_accents(authors_str)
        authors_str = re.sub("[^a-zA-Z, ]+", "", authors_str)

        match_count = 0
        for author_name in authors_str.split(" and "):
            # Only check last names because first names amy be abbreviated
            author_name = author_name.split(",")[0].replace(" ", "")
            if author_name in text:
                match_count += 1

        nr_authors = len(authors_str.split(" and "))
        if match_count / nr_authors > 0.8:
            return True

        return False


def register(quality_model: colrev.record.qm.quality_model.QualityModel) -> None:
    """Register the checker"""
    quality_model.register_checker(AuthorNotInPDFChecker(quality_model))



#! /usr/bin/env python
"""SearchSource: DBLP"""
from __future__ import annotations

import re
import typing
from multiprocessing import Lock
from pathlib import Path

import requests
import zope.interface
from pydantic import BaseModel
from pydantic import Field

import colrev.exceptions as colrev_exceptions
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
import colrev.record.record_prep
import colrev.record.record_similarity
import colrev.settings
from colrev.constants import Fields
from colrev.constants import FieldValues
from colrev.constants import RecordState
from colrev.constants import SearchSourceHeuristicStatus
from colrev.constants import SearchType
from colrev.packages.dblp.src import dblp_api

# pylint: disable=unused-argument
# pylint: disable=duplicate-code


class DBLPSearchSourceSettings(colrev.settings.SearchSource, BaseModel):
    """Settings for DBLPSearchSource"""

    # pylint: disable=duplicate-code
    # pylint: disable=too-many-instance-attributes
    endpoint: str
    filename: Path
    search_type: SearchType
    search_parameters: dict
    comment: typing.Optional[str]

    _details = {
        "search_parameters": {
            "tooltip": "Currently supports a scope item "
            "with venue_key and journal_abbreviated fields."
        },
    }


@zope.interface.implementer(colrev.package_manager.interfaces.SearchSourceInterface)
class DBLPSearchSource:
    """DBLP API"""

    source_identifier = "dblp_key"
    search_types = [
        SearchType.API,
        SearchType.MD,
        SearchType.TOC,
    ]
    endpoint = "colrev.dblp"

    ci_supported: bool = Field(default=True)
    heuristic_status = SearchSourceHeuristicStatus.supported

    settings_class = DBLPSearchSourceSettings

    _timeout: int = 20

    def __init__(
        self,
        *,
        source_operation: colrev.process.operation.Operation,
        settings: typing.Optional[dict] = None,
    ) -> None:
        self.review_manager = source_operation.review_manager
        self.search_source = self._get_search_source(settings)
        self.dblp_lock = Lock()
        self.origin_prefix = self.search_source.get_origin_prefix()
        _, self.email = self.review_manager.get_committer()

    def _get_search_source(
        self, settings: typing.Optional[dict]
    ) -> colrev.settings.SearchSource:

        # DBLP as a search_source
        if settings:
            return self.settings_class(**settings)

        # DBLP as an md-prep source
        dblp_md_filename = Path("data/search/md_dblp.bib")
        dblp_md_source_l = [
            s
            for s in self.review_manager.settings.sources
            if s.filename == dblp_md_filename
        ]
        if dblp_md_source_l:
            return dblp_md_source_l[0]

        return colrev.settings.SearchSource(
            endpoint=self.endpoint,
            filename=dblp_md_filename,
            search_type=SearchType.MD,
            search_parameters={},
            comment="",
        )

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for DBLP"""

        result = {"confidence": 0.0}
        # Simple heuristic:
        # pylint: disable=colrev-missed-constant-usage
        if "dblp_key" in data:
            result["confidence"] = 1.0
            return result

        if "dblp computer science bibliography" in data:
            if data.count("dblp computer science bibliography") >= data.count("\n@"):
                result["confidence"] = 1.0

        return result

    @classmethod
    def add_endpoint(
        cls,
        operation: colrev.ops.search.Search,
        params: str,
    ) -> colrev.settings.SearchSource:
        """Add SearchSource as an endpoint (based on query provided to colrev search --add )"""

        params_dict = {}
        if params:
            if params.startswith("http"):
                params_dict = {Fields.URL: params}
            else:
                for item in params.split(";"):
                    key, value = item.split("=")
                    params_dict[key] = value

        search_type = operation.select_search_type(
            search_types=cls.search_types, params=params_dict
        )

        if search_type == SearchType.API:
            if len(params_dict) == 0:
                search_source = operation.create_api_source(endpoint=cls.endpoint)

            # pylint: disable=colrev-missed-constant-usage
            elif "url" in params_dict:
                api_url = "https://dblp.org/search/publ/api?q="
                query = (
                    params_dict["url"]
                    .replace("https://dblp.org/search?q=", api_url)
                    .replace("https://dblp.org/search/publ?q=", api_url)
                )

                filename = operation.get_unique_filename(file_path_string="dblp")
                search_source = colrev.settings.SearchSource(
                    endpoint=cls.endpoint,
                    filename=filename,
                    search_type=SearchType.API,
                    search_parameters={"query": query},
                    comment="",
                )
            else:
                raise NotImplementedError

        # elif search_type == SearchType.TOC:
        else:
            raise colrev_exceptions.PackageParameterError(
                f"Cannot add dblp endpoint with query {params}"
            )

        operation.add_source_and_search(search_source)
        return search_source

    def check_availability(
        self, *, source_operation: colrev.process.operation.Operation
    ) -> None:
        """Check status (availability) of DBLP API"""
        if self.review_manager.force_mode:
            return

        api = dblp_api.DBLPAPI(
            params={},
            email=self.email,
            session=self.review_manager.get_cached_session(),
            timeout=self._timeout,
        )
        api.check_availability()

    def _run_md_search(
        self,
        *,
        dblp_feed: colrev.ops.search_api_feed.SearchAPIFeed,
    ) -> None:

        api = dblp_api.DBLPAPI(
            params={},
            email=self.email,
            session=self.review_manager.get_cached_session(),
            rerun=True,
            timeout=self._timeout,
        )

        for feed_record_dict in dblp_feed.feed_records.values():
            if Fields.TITLE not in feed_record_dict:
                continue
            api.params = {"query": feed_record_dict[Fields.TITLE]}
            api.set_url_from_query()
            for retrieved_record in api.retrieve_records():
                try:
                    if (
                        retrieved_record.data["dblp_key"]
                        != feed_record_dict["dblp_key"]
                    ) or retrieved_record.data.get("type", "") == "Editorship":
                        continue

                    dblp_feed.add_update_record(retrieved_record)
                except colrev_exceptions.NotFeedIdentifiableException:
                    continue

        dblp_feed.save()

    def _run_api_search(
        self,
        *,
        dblp_feed: colrev.ops.search_api_feed.SearchAPIFeed,
        rerun: bool,
    ) -> None:

        api = dblp_api.DBLPAPI(
            params=self.search_source.search_parameters,
            email=self.email,
            session=self.review_manager.get_cached_session(),
            rerun=(len(dblp_feed.feed_records) < 100 or rerun),
            timeout=self._timeout,
        )

        total = api.total
        self.review_manager.logger.info(f"Total: {total:,}")
        if not rerun and len(dblp_feed.feed_records) > 0:
            self.review_manager.logger.info(
                "Retrieving latest results (no estimate available)"
            )
        elif total > 0 and rerun:
            seconds = 10 + (total / 10)
            if total > 1000:
                seconds += total / 100 * 60  # request timeouts

            formatted_time = (
                str(int(seconds // 3600)).zfill(2)
                + ":"
                + str(int((seconds % 3600) // 60)).zfill(2)
                + ":"
                + str(int(seconds % 60)).zfill(2)
            )

            self.review_manager.logger.info(
                f"Estimated runtime [hh:mm:ss]: {formatted_time}"
            )

        while True:
            api.set_next_url()

            for retrieved_record in api.retrieve_records():
                try:

                    if "scope" in self.search_source.search_parameters and (
                        f"{self.search_source.search_parameters['scope']['venue_key']}/"
                        not in retrieved_record.data["dblp_key"]
                        or retrieved_record.data.get(Fields.ENTRYTYPE, "")
                        not in [
                            "article",
                            "inproceedings",
                        ]
                    ):
                        continue

                    dblp_feed.add_update_record(retrieved_record)

                except colrev_exceptions.NotFeedIdentifiableException as exc:
                    print(exc)
                    continue

            if api.processed_all_urls():
                break

        dblp_feed.save()

    def _validate_source(self) -> None:
        """Validate the SearchSource (parameters etc.)"""
        source = self.search_source
        self.review_manager.logger.debug(f"Validate SearchSource {source.filename}")

        assert source.search_type in self.search_types

        # maybe : validate/assert that the venue_key is available
        if source.search_type == SearchType.TOC:
            assert "scope" in source.search_parameters
            if "venue_key" not in source.search_parameters["scope"]:
                raise colrev_exceptions.InvalidQueryException(
                    "venue_key required in search_parameters/scope"
                )
            if "journal_abbreviated" not in source.search_parameters["scope"]:
                raise colrev_exceptions.InvalidQueryException(
                    "journal_abbreviated required in search_parameters/scope"
                )
        elif source.search_type == SearchType.API:
            assert "query" in source.search_parameters

        elif source.search_type == SearchType.MD:
            pass  # No parameters required
        else:
            raise colrev_exceptions.InvalidQueryException(
                "scope or query required in search_parameters"
            )

        self.review_manager.logger.debug(f"SearchSource {source.filename} validated")

    def search(self, rerun: bool) -> None:
        """Run a search of DBLP"""

        self._validate_source()

        dblp_feed = self.search_source.get_api_feed(
            review_manager=self.review_manager,
            source_identifier=self.source_identifier,
            update_only=(not rerun),
        )
        if self.search_source.search_type == SearchType.MD:
            self._run_md_search(dblp_feed=dblp_feed)

        elif self.search_source.search_type in [
            SearchType.API,
            SearchType.TOC,
        ]:
            self._run_api_search(dblp_feed=dblp_feed, rerun=rerun)

        else:
            raise NotImplementedError

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".bib":

            def field_mapper(record_dict: dict) -> None:
                if "timestamp" in record_dict:
                    record_dict[f"{self.endpoint}.timestamp"] = record_dict.pop(
                        "timestamp"
                    )
                if "biburl" in record_dict:
                    record_dict[f"{self.endpoint}.biburl"] = record_dict.pop("biburl")
                if "bibsource" in record_dict:
                    record_dict[f"{self.endpoint}.bibsource"] = record_dict.pop(
                        "bibsource"
                    )
                if "dblp_key" in record_dict:
                    record_dict[Fields.DBLP_KEY] = record_dict.pop("dblp_key")

            records = colrev.loader.load_utils.load(
                filename=self.search_source.filename,
                unique_id_field="ID",
                field_mapper=field_mapper,
                logger=self.review_manager.logger,
            )
            return records

        raise NotImplementedError

    def prepare(
        self,
        record: colrev.record.record_prep.PrepRecord,
        source: colrev.settings.SearchSource,
    ) -> colrev.record.record_prep.PrepRecord:
        """Source-specific preparation for DBLP"""

        if record.data.get(Fields.AUTHOR, FieldValues.UNKNOWN) != FieldValues.UNKNOWN:
            # DBLP appends identifiers to non-unique authors
            record.update_field(
                key=Fields.AUTHOR,
                value=str(re.sub(r"[0-9]{4}", "", record.data[Fields.AUTHOR])),
                source="dblp",
                keep_source_if_equal=True,
            )
        record.remove_field(key="colrev.dblp.bibsource")
        if any(x in record.data.get(Fields.URL, "") for x in ["dblp.org", "doi.org"]):
            record.remove_field(key=Fields.URL)
        record.remove_field(key="colrev.dblp.timestamp")

        return record

    def prep_link_md(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.record.Record,
        save_feed: bool = True,
        timeout: int = 60,
    ) -> colrev.record.record.Record:
        """Retrieve masterdata from DBLP based on similarity with the record provided"""

        if any(self.origin_prefix in o for o in record.data[Fields.ORIGIN]):
            # Already linked to a crossref record
            return record
        if Fields.TITLE not in record.data:
            return record

        try:

            # Note: queries combining title+author/journal do not seem to work any more
            api = dblp_api.DBLPAPI(
                params={"query": record.data[Fields.TITLE]},
                email=self.email,
                session=self.review_manager.get_cached_session(),
                timeout=timeout,
            )

            retrieved_records = api.retrieve_records()
            if not retrieved_records:
                return record
            retrieved_record = retrieved_records[0]
            if Fields.DBLP_KEY in record.data:
                if retrieved_record.data["dblp_key"] != record.data[Fields.DBLP_KEY]:
                    return record

            if not colrev.record.record_similarity.matches(record, retrieved_record):
                return record

            try:
                self.dblp_lock.acquire(timeout=60)

                # Note : need to reload file
                # because the object is not shared between processes
                dblp_feed = self.search_source.get_api_feed(
                    review_manager=self.review_manager,
                    source_identifier=self.source_identifier,
                    update_only=False,
                    prep_mode=True,
                )

                dblp_feed.add_update_record(retrieved_record)

                # Assign schema
                retrieved_record.data[Fields.DBLP_KEY] = retrieved_record.data.pop(
                    "dblp_key"
                )

                record.merge(
                    retrieved_record,
                    default_source=retrieved_record.data[Fields.ORIGIN][0],
                )
                record.set_masterdata_complete(
                    source=retrieved_record.data[Fields.ORIGIN][0],
                    masterdata_repository=self.review_manager.settings.is_curated_repo(),
                )
                record.set_status(RecordState.md_prepared)
                if "Withdrawn (according to DBLP)" in record.data.get(
                    "colrev.dblp.warning", ""
                ):
                    record.prescreen_exclude(reason=FieldValues.RETRACTED)
                    # record.remove_field(key="warning")

                dblp_feed.save()
                self.dblp_lock.release()

            except (colrev_exceptions.NotFeedIdentifiableException,):
                self.dblp_lock.release()

        except requests.exceptions.RequestException:
            pass
        except colrev_exceptions.ServiceNotAvailableException:
            if self.review_manager.force_mode:
                self.review_manager.logger.error("Service not available: DBLP")

        return record



#! /usr/bin/env python
"""CoLRev init operation: Create a project and specify settings."""
from __future__ import annotations

import json
import logging
import os
import platform
import shutil
from importlib.metadata import version
from pathlib import Path
from subprocess import CalledProcessError  # nosec
from subprocess import check_call  # nosec
from subprocess import DEVNULL  # nosec
from subprocess import STDOUT  # nosec

import git
from git.exc import InvalidGitRepositoryError

import colrev.env.docker_manager
import colrev.env.environment_manager
import colrev.env.utils
import colrev.exceptions as colrev_exceptions
import colrev.ops.check
import colrev.package_manager.package_manager
import colrev.review_manager
import colrev.settings
from colrev.constants import Colors
from colrev.constants import EndpointType
from colrev.constants import Fields

# pylint: disable=too-few-public-methods


class Initializer:
    """Initialize a CoLRev project"""

    review_manager: colrev.review_manager.ReviewManager

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        *,
        review_type: str,
        target_path: Path,
        example: bool = False,
        force_mode: bool = False,
        light: bool = False,
        exact_call: str = "",
    ) -> None:
        p_man = colrev.package_manager.package_manager.PackageManager()
        self.review_type = self._format_review_type(review_type)
        if not p_man.is_installed(self.review_type):
            print(
                f"Run {Colors.ORANGE}colrev install all_internal_packages{Colors.END} "
                "before colrev init"
            )
            raise colrev.exceptions.MissingDependencyError(self.review_type)
        self.force_mode = force_mode
        self.target_path = target_path
        os.chdir(target_path)
        self.review_manager = colrev.review_manager.ReviewManager(
            path_str=str(self.target_path), force_mode=True, navigate_to_home_dir=False
        )
        self.title = str(self.target_path.name)
        self._setup_repo(
            example=example,
            exact_call=exact_call,
            no_docker=light,
        )

        # Install packages
        p_man.install_project(review_manager=self.review_manager, force_reinstall=False)

    def _setup_repo(
        self,
        *,
        example: bool,
        exact_call: str,
        no_docker: bool,
    ) -> None:
        self.no_docker = no_docker
        # check whether docker is available
        try:
            colrev.env.docker_manager.DockerManager.check_docker_installed()
        except colrev_exceptions.MissingDependencyError:
            self.no_docker = True

        self._reset_if_existing_repo_with_single_commit()
        self._check_init_precondition()
        self.review_manager.logger.info("Create CoLRev repository")
        self._setup_git()
        self._setup_files()
        self._setup_settings()
        self._finalize()
        if example:
            self._create_example_repo()

        self.review_manager = colrev.review_manager.ReviewManager(exact_call=exact_call)

        self._git_rm_settings()

        self.review_manager.dataset.create_commit(
            msg="Init: Create CoLRev repository",
            manual_author=True,
            skip_hooks=True,
        )
        self._add_colrev_project_details()
        self.review_manager.dataset.create_commit(
            msg="Init: Create CoLRev project",
            manual_author=True,
            skip_hooks=True,
        )

        self._register_repo(example=example)

        self._post_commit_edits()

        self.review_manager.logger.info(
            "%sCompleted init operation%s", Colors.GREEN, Colors.END
        )

    def _git_rm_settings(self) -> None:
        colrev.ops.check.CheckOperation(self.review_manager)
        git_repo = self.review_manager.dataset.get_repo()
        git_repo.index.remove([self.review_manager.paths.settings])

    def _add_colrev_project_details(self) -> None:

        # Note : to avoid file setup at colrev status (calls data_operation.main)
        data_operation = self.review_manager.get_data_operation(
            notify_state_transition_operation=False
        )
        data_operation.main(silent_mode=True)

        git_repo = self.review_manager.dataset.get_repo()
        git_repo.git.add(all=True)

    def _format_review_type(self, review_type: str) -> str:
        formatted_review_type = review_type.replace("-", "_").lower().replace(" ", "_")
        if "." not in formatted_review_type:
            formatted_review_type = "colrev." + formatted_review_type

        return formatted_review_type

    def _reset_if_existing_repo_with_single_commit(self) -> None:
        def is_empty_colrev_template() -> bool:
            files = list(
                os.path.join(root, file)
                for root, _, files in os.walk(self.target_path)
                for file in files
            )
            files = [os.path.relpath(file, self.target_path) for file in files]
            files = [
                file
                for file in files
                if not file.startswith(".git/")
                and not file.startswith(".devcontainer/")
            ]
            return not files

        try:
            git_repo = git.Repo.init()
            if len(list(git_repo.iter_commits())) == 1:
                if is_empty_colrev_template():
                    return
                print("Detected existing repository")
                if "y" != input("Reset existing repository? (y/n) "):
                    raise colrev_exceptions.CoLRevException("Operation aborted.")
                for root, dirs, files in os.walk(self.target_path):
                    for file in files:
                        os.remove(os.path.join(root, file))
                    for directory in dirs:
                        shutil.rmtree(os.path.join(root, directory))
        except (InvalidGitRepositoryError, ValueError):
            pass

    def _check_init_precondition(self) -> None:
        if self.force_mode:
            return
        cur_content = [
            str(x.relative_to(self.target_path)) for x in self.target_path.glob("**/*")
        ]
        cur_content = [
            x for x in cur_content if not x.startswith("venv") and x != ".history"
        ]
        if str(self.review_manager.paths.REPORT_FILE) in cur_content:
            cur_content.remove(str(self.review_manager.paths.REPORT_FILE))

        if all(x.startswith((".git", ".devcontainer", ".vscode")) for x in cur_content):
            return

        if all(x.startswith((".git", ".devcontainer", ".vscode")) for x in cur_content):
            return

        if cur_content:
            raise colrev_exceptions.NonEmptyDirectoryError(
                filepath=self.target_path, content=cur_content
            )

        environment_manager = colrev.env.environment_manager.EnvironmentManager()
        environment_manager.get_name_mail_from_git()

        try:
            colrev.env.docker_manager.DockerManager.check_docker_installed()
        except colrev_exceptions.MissingDependencyError as exc:  # pragma: no cover
            if not self.no_docker:
                raise colrev_exceptions.CoLRevException(
                    "Docker not installed. Docker is optional but recommended.\n"
                    "For more information, see "
                    "https://colrev-environment.github.io/colrev/manual/manual.html"
                    "To init a repository without Docker, run "
                    f"{Colors.ORANGE}colrev init --light{Colors.END}"
                ) from exc

    def _setup_git(self) -> None:
        self.review_manager.logger.info("Set up git repository")

        git.Repo.init()

        # To check if git actors are set
        environment_manager = colrev.env.environment_manager.EnvironmentManager()
        environment_manager.get_name_mail_from_git()

        logging.info("Install latest pre-commmit hooks")
        scripts_to_call = [
            {
                "description": "Install pre-commit hooks",
                "command": ["pre-commit", "install"],
            },
            {
                "description": "",
                "command": [
                    "pre-commit",
                    "install",
                    "--hook-type",
                    "prepare-commit-msg",
                ],
            },
            {
                "description": "",
                "command": ["pre-commit", "install", "--hook-type", "pre-push"],
            },
            {"description": "", "command": ["pre-commit", "autoupdate"]},
        ]
        for script_to_call in scripts_to_call:
            try:
                if script_to_call["description"]:
                    self.review_manager.logger.debug(
                        "%s...", script_to_call["description"]
                    )
                check_call(
                    script_to_call["command"], stdout=DEVNULL, stderr=STDOUT
                )  # nosec
            except CalledProcessError:
                if " ".join(script_to_call["command"]) == "pre-commit autoupdate":
                    pass
                else:
                    self.review_manager.logger.error(
                        "%sFailed: %s%s",
                        Colors.RED,
                        " ".join(script_to_call),
                        Colors.END,
                    )

    def _fix_pre_commit_hooks_windows(self) -> None:
        # https://stackoverflow.com/questions/12410164/github-for-windows-pre-commit-hook
        if any(platform.win32_ver()):
            with open(".git/hooks/pre-commit", encoding="utf-8") as file:
                lines = file.readlines()

            lines[0] = """#!C:/Program\\ Files/Git/usr/bin/sh.exe\n"""

            with open(".git/hooks/pre-commit", "w", encoding="utf-8") as file:
                file.writelines(lines)

    def _setup_files(self) -> None:

        # Note: parse instead of copy to avoid format changes
        settings_filedata = colrev.env.utils.get_package_file_content(
            module="colrev.ops", filename=Path("init/settings.json")
        )
        if settings_filedata:
            settings = json.loads(settings_filedata.decode("utf-8"))
            settings["project"]["review_type"] = str(self.review_type)
            with open(self.review_manager.paths.settings, "w", encoding="utf8") as file:
                json.dump(settings, file, indent=4)

        self.review_manager.paths.search.mkdir(parents=True)
        self.review_manager.paths.pdf.mkdir(parents=True)

        colrev_path = Path.home() / Path("colrev")
        colrev_path.mkdir(exist_ok=True, parents=True)

        files_to_retrieve = [
            [Path("ops/init/readme.md"), self.review_manager.paths.readme],
            [
                Path("ops/init/pre-commit-config.yaml"),
                self.review_manager.paths.pre_commit_config,
            ],
            [
                Path("ops/init/markdownlint.yaml"),
                self.review_manager.path / Path(".markdownlint.yaml"),
            ],
            [
                Path("ops/init/pre-commit.yml"),
                self.review_manager.path / Path(".github/workflows/pre-commit.yml"),
            ],
            [
                Path("ops/init/gitattributes"),
                self.review_manager.path / Path(".gitattributes"),
            ],
            [
                Path("ops/init/LICENSE-CC-BY-4.0.txt"),
                self.review_manager.path / Path("LICENSE.txt"),
            ],
            [
                Path("ops/init/colrev_update.yml"),
                self.review_manager.path / Path(".github/workflows/colrev_update.yml"),
            ],
        ]
        for retrieval_path, target_path in files_to_retrieve:
            colrev.env.utils.retrieve_package_file(
                template_file=retrieval_path, target=target_path
            )

    def _setup_settings(self) -> None:

        self.review_manager = colrev.review_manager.ReviewManager()
        settings = self.review_manager.settings

        committer, email = self.review_manager.get_committer()
        settings.project.authors = [
            colrev.settings.Author(
                name=committer,
                initials="".join(part[0] for part in committer.split(" ")),
                email=email,
            )
        ]

        colrev_version = version("colrev")
        if "+" in colrev_version:
            colrev_version = colrev_version[: colrev_version.rfind("+")]
        settings.project.colrev_version = colrev_version

        settings.project.title = self.title
        self.review_type = settings.project.review_type

        # Principle: adapt values provided by the default SETTINGS_FILE
        # instead of creating a new SETTINGS_FILE
        package_manager = self.review_manager.get_package_manager()
        review_type_class = package_manager.get_package_endpoint_class(
            package_type=EndpointType.review_type,
            package_identifier=self.review_type,
        )
        check_operation = colrev.ops.check.CheckOperation(self.review_manager)
        review_type_object = review_type_class(
            operation=check_operation,
            settings={"endpoint": settings.project.review_type},
        )

        settings = review_type_object.initialize(settings=settings)
        self.review_manager.save_settings()

        project_title = self.review_manager.settings.project.title
        if "review" in project_title.lower():
            colrev.env.utils.inplace_change(
                filename=Path("readme.md"),
                old_string="{{project_title}}",
                new_string=project_title.rstrip(" ").capitalize(),
            )
        else:
            package_manager = self.review_manager.get_package_manager()
            r_type_suffix = str(review_type_object)

            colrev.env.utils.inplace_change(
                filename=Path("readme.md"),
                old_string="{{project_title}}",
                new_string=project_title.rstrip(" ").capitalize()
                + f": A {r_type_suffix} protocol",
            )

        if self.no_docker:
            settings.data.data_package_endpoints = [
                x
                for x in settings.data.data_package_endpoints
                if x["endpoint"] not in ["colrev.paper_md"]
            ]
            settings.sources = [
                x for x in settings.sources if x.endpoint not in ["colrev.files_dir"]
            ]

            settings.pdf_prep.pdf_prep_package_endpoints = [
                x
                for x in settings.pdf_prep.pdf_prep_package_endpoints
                if x["endpoint"]
                not in [
                    "colrev.ocrmypdf",
                    "colrev.remove_coverpage",
                    "colrev.remove_last_page",
                    "colrev.grobid_tei",
                ]
            ]

        self.review_manager.save_settings()

    def _finalize(self) -> None:
        settings = self.review_manager.settings

        self.review_manager.logger.info("Set up %s", self.review_type)

        for source in settings.sources:
            self.review_manager.logger.info(
                " add search %s", source.endpoint.replace("colrev.", "")
            )

        for data_package_endpoint in settings.data.data_package_endpoints:
            self.review_manager.logger.info(
                " add data   %s",
                data_package_endpoint["endpoint"].replace("colrev.", ""),
            )

        with open(
            self.review_manager.paths.records, mode="w", encoding="utf-8"
        ) as file:
            file.write("\n")

        self._fix_pre_commit_hooks_windows()

        git_repo = self.review_manager.dataset.get_repo()
        git_repo.git.add(all=True)

    def _register_repo(self, *, example: bool) -> None:
        if example or "pytest" in os.getcwd():
            return
        self.review_manager.logger.info("Register CoLRev repository")
        environment_manager = self.review_manager.get_environment_manager()
        environment_manager.register_repo(
            self.target_path, logger=self.review_manager.logger
        )

    def _post_commit_edits(self) -> None:
        if self.review_type != "colrev.curated_masterdata":
            return

        self.review_manager.logger.info("Post-commit edits")
        self.review_manager.settings.project.curation_url = "TODO"
        self.review_manager.settings.project.curated_fields = [
            Fields.URL,
            Fields.DOI,
            "TODO",
        ]

        pdf_source_l = [
            s
            for s in self.review_manager.settings.sources
            if "data/search/pdfs.bib" == str(s.filename)
        ]
        if pdf_source_l:
            pdf_source = pdf_source_l[0]
            pdf_source.search_parameters = {
                "scope": {
                    "path": "pdfs",
                    Fields.JOURNAL: "TODO",
                    "subdir_pattern": "TODO:volume_number|year",
                }
            }

        crossref_source_l = [
            s
            for s in self.review_manager.settings.sources
            if "data/search/CROSSREF.bib" == str(s.filename)
        ]
        if crossref_source_l:
            crossref_source = crossref_source_l[0]
            crossref_source.search_parameters = {"scope": {"journal_issn": "TODO"}}

        self.review_manager.save_settings()
        self.review_manager.logger.info("Completed setup.")

    def _create_example_repo(self) -> None:
        """The example repository is intended to provide an initial illustration
        of CoLRev. It focuses on a quick overview of the process and does
        not cover advanced features or special cases."""

        self.review_manager.logger.info("Include 30_example_records.bib")
        colrev.env.utils.retrieve_package_file(
            template_file=Path("ops/init/30_example_records.bib"),
            target=Path("data/search/30_example_records.bib"),
        )

        git_repo = self.review_manager.dataset.get_repo()
        git_repo.index.add(["data/search/30_example_records.bib"])

        with open(self.review_manager.paths.settings, encoding="utf-8") as file:
            settings = json.load(file)

        settings["dedupe"]["dedupe_package_endpoints"] = [{"endpoint": "colrev.dedupe"}]
        settings["sources"] = [
            {
                "endpoint": "colrev.unknown_source",
                "filename": str(Path("data/search/30_example_records.bib")),
                "search_type": "DB",
                "search_parameters": {
                    "query_file": str(Path("data/search/30_example_records_query.txt"))
                },
                "comment": "",
            }
        ]

        with open(self.review_manager.paths.settings, "w", encoding="utf-8") as outfile:
            json.dump(settings, outfile, indent=4)
        git_repo.index.add([self.review_manager.paths.SETTINGS_FILE])



#! /usr/bin/env python
"""Adding of journal rankings to metadata"""
from __future__ import annotations

import zope.interface
from pydantic import Field

import colrev.env.local_index
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import Fields

# pylint: disable=too-few-public-methods


@zope.interface.implementer(colrev.package_manager.interfaces.PrepInterface)
class AddJournalRanking:
    """Prepares records based on journal rankings"""

    settings_class = colrev.package_manager.package_settings.DefaultSettings
    source_correction_hint = "check with the developer"
    always_apply_changes = False
    ci_supported: bool = Field(default=False)

    def __init__(
        self,
        *,
        prep_operation: colrev.ops.prep.Prep,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        self.settings = self.settings_class(**settings)
        self.local_index = colrev.env.local_index.LocalIndex()

    def prepare(
        self,
        record: colrev.record.record_prep.PrepRecord,
    ) -> colrev.record.record.Record:
        """Add Journalranking to Metadata"""

        if record.data.get(Fields.JOURNAL, "") == "":
            return record

        rankings = self.local_index.get_journal_rankings(record.data[Fields.JOURNAL])
        # extend: include journal-impact factor or ranking category
        if rankings:
            rankings_str = ",".join(r["ranking"] for r in rankings)
        else:
            rankings_str = "not included in a ranking"

        record.update_field(
            key="journal_ranking",
            value=rankings_str,
            source="add_journal_ranking",
            note="",
        )

        return record



#! /usr/bin/env python
"""Checker for mostly-all-caps fields."""
from __future__ import annotations

import colrev.env.utils
import colrev.record.qm.quality_model
from colrev.constants import DefectCodes
from colrev.constants import Fields
from colrev.constants import FieldValues

# pylint: disable=too-few-public-methods


class MostlyAllCapsFieldChecker:
    """The MostlyAllCapsFieldChecker"""

    msg = DefectCodes.MOSTLY_ALL_CAPS

    def __init__(
        self, quality_model: colrev.record.qm.quality_model.QualityModel
    ) -> None:
        self.quality_model = quality_model

    def run(self, *, record: colrev.record.record.Record) -> None:
        """Run the mostly-all-caps checks"""
        for key in [
            Fields.AUTHOR,
            Fields.TITLE,
            Fields.JOURNAL,
            Fields.BOOKTITLE,
            Fields.EDITOR,
        ]:
            if (
                key not in record.data
                or record.ignored_defect(key=key, defect=self.msg)
                or record.data[key] == FieldValues.UNKNOWN
            ):
                continue

            if record.masterdata_is_curated():
                continue

            if self._is_mostly_all_caps(record=record, key=key):
                record.add_field_provenance_note(key=key, note=self.msg)
            else:
                record.remove_field_provenance_note(key=key, note=self.msg)

    def _is_mostly_all_caps(
        self, *, record: colrev.record.record.Record, key: str
    ) -> bool:
        """Check if the field is mostly all caps"""

        # Online sources/software can be short/have caps
        if (
            record.data.get(Fields.ENTRYTYPE, "") == "online"
            and key == Fields.TITLE
            and len(record.data[Fields.TITLE]) < 10
        ):
            return False

        if (
            colrev.env.utils.percent_upper_chars(record.data[key].replace(" and ", ""))
            < 0.7
        ):
            return False

        # container-title-abbreviated
        if key in [Fields.JOURNAL, Fields.BOOKTITLE] and len(record.data[key]) < 6:
            return False

        if record.data[key].upper() == "PLOS ONE":
            return False

        return True


def register(quality_model: colrev.record.qm.quality_model.QualityModel) -> None:
    """Register the checker"""
    quality_model.register_checker(MostlyAllCapsFieldChecker(quality_model))



#! /usr/bin/env python
"""Discovering and using packages."""
from __future__ import annotations

import importlib.metadata
import importlib.util
import json
import platform
import subprocess
import sys
import typing
from typing import Any

import colrev.exceptions as colrev_exceptions
import colrev.package_manager.colrev_internal_packages
import colrev.package_manager.package
from colrev.constants import Colors
from colrev.constants import EndpointType
from colrev.constants import Filepaths


class PackageManager:
    """The PackageManager provides functionality for package lookup and discovery"""

    def _get_package_identifiers(self) -> list:
        group = "colrev"
        return [
            dist.metadata["name"]
            for dist in importlib.metadata.distributions()
            for ep in dist.entry_points
            if ep.group == group
        ]

    def _load_type_identifier_endpoint_dict(self) -> dict:
        type_identifier_endpoint_dict: typing.Dict[
            EndpointType, typing.Dict[str, Any]
        ] = {endpoint_type: {} for endpoint_type in EndpointType}

        for package_identifier in self._get_package_identifiers():
            try:
                package = colrev.package_manager.package.Package(package_identifier)
                package.add_to_type_identifier_endpoint_dict(
                    type_identifier_endpoint_dict
                )
            except colrev_exceptions.MissingDependencyError as exc:
                print(exc)

        return type_identifier_endpoint_dict

    def discover_packages(self, *, package_type: EndpointType) -> typing.Dict:
        """Discover packages (registered in the CoLRev environment)"""

        # [{'package_endpoint_identifier': 'colrev.abi_inform_proquest',
        #   'status': '|EXPERIMENTAL|',
        #   'short_description': 'ABI/INFORM (ProQuest) ...'},
        #  ...]
        with open(Filepaths.PACKAGES_ENDPOINTS_JSON, encoding="utf-8") as file:
            package_endpoints = json.load(file)
            return package_endpoints[package_type.value]

    def discover_installed_packages(self, *, package_type: EndpointType) -> typing.Dict:
        """Discover installed packages"""

        # {EndpointType.review_type:
        #   {'colrev.blank': {'endpoint': 'colrev.packages.review_types.blank.BlankReview'},
        #     ...
        # }

        type_identifier_endpoint_dict = self._load_type_identifier_endpoint_dict()
        return type_identifier_endpoint_dict[package_type]

    def get_package_endpoint_class(  # type: ignore
        self, *, package_type: EndpointType, package_identifier: str
    ):
        """Load a package endpoint"""

        package = colrev.package_manager.package.Package(package_identifier)
        return package.get_endpoint_class(package_type)

    def is_installed(self, package_name: str) -> bool:
        """Check if a package is installed"""

        # Notes:
        # Cannot import directly and check whether it fails
        # because the package format is non-standard (".")
        # We deactivate the is_installed() temporarily
        # until internal colrev packages comply with naming conventions.
        if platform.system() in ["Darwin", "Windows", "Linux"]:
            return True  # Return True for macOS, Linux, and Windows
        print(package_name)

        # try:
        #     if sys.version_info >= (3, 10):
        #         # Use packages_distributions in Python 3.10+
        #         from importlib.metadata import packages_distributions

        #         installed_packages = packages_distributions()

        #         if package_name.replace("-", "_") in installed_packages:
        #             return True
        #         if (
        #             "src" in installed_packages
        #             and package_name.replace("-", "_") in installed_packages["src"]
        #         ):
        #             return True

        #         return False
        #     else:
        #         # Fallback for Python < 3.10 using the distribution method
        #         importlib.metadata.distribution(package_name.replace("-", "_"))
        #         return True
        # except PackageNotFoundError:
        #     return False
        # except importlib.metadata.PackageNotFoundError:
        #     return False
        return True

    def _get_packages_to_install(
        self,
        *,
        review_manager: colrev.review_manager.ReviewManager,
        force_reinstall: bool,
    ) -> typing.List[str]:

        review_manager.logger.info("Packages:")
        packages = review_manager.settings.get_packages()

        installed_packages = []
        for package in packages:
            if self.is_installed(package):
                installed_packages.append(package)
                review_manager.logger.info(
                    f" {Colors.GREEN}{package}: installed{Colors.END}"
                )
            else:
                review_manager.logger.info(
                    f" {Colors.ORANGE}{package}: not installed{Colors.END}"
                )

        if not force_reinstall:
            packages = [p for p in packages if p not in installed_packages]

        return packages

    def install_project(
        self,
        *,
        review_manager: colrev.review_manager.ReviewManager,
        force_reinstall: bool,
    ) -> None:
        """Install all packages required for the CoLRev project"""

        review_manager.logger.info("Install project")
        packages = self._get_packages_to_install(
            review_manager=review_manager, force_reinstall=force_reinstall
        )
        if len(packages) == 0:
            review_manager.logger.info("All packages are already installed")
            return

        self.install(packages=packages)

    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-branches
    def install(
        self,
        *,
        packages: typing.List[str],
        upgrade: bool = True,
        editable: str = "",
        force_reinstall: bool = True,
        no_cache_dir: bool = True,
    ) -> None:
        """Install packages"""

        internal_packages_dict = (
            colrev.package_manager.colrev_internal_packages.get_internal_packages_dict()
        )

        if len(packages) == 1 and packages[0] == "all_internal_packages":
            packages = list(internal_packages_dict.keys())

        # Install packages from colrev monorepository first
        colrev_packages = []
        for package in packages:
            if package in internal_packages_dict:
                colrev_packages.append(package)
        packages = [p for p in packages if p not in colrev_packages]

        print(f"ColRev packages: {colrev_packages}")
        if colrev_packages:

            colrev_package_paths = [
                p_path
                for p_name, p_path in internal_packages_dict.items()
                if p_name in colrev_packages
            ]
            args = [sys.executable, "-m", "pip", "install"]
            args += colrev_package_paths
            if upgrade:
                args += ["--upgrade"]
            if editable:
                args += ["--editable", editable]
            if force_reinstall:
                args += ["--force-reinstall"]
            if no_cache_dir:
                args += ["--no-cache-dir"]
            # sys.argv = args
            # run_module("pip", run_name="__main__")
            # use subprocess because run_module does not return control
            subprocess.run(args, check=False)

        print(f"Other packages: {packages}")
        if packages:
            args = [sys.executable, "-m", "pip", "install"]
            if upgrade:
                args += ["--upgrade"]
            if editable:
                args += ["--editable", editable]
            if force_reinstall:
                args += ["--force-reinstall"]
            if no_cache_dir:
                args += ["--no-cache-dir"]
            args += list(packages)
            # sys.argv = args
            # run_module("pip", run_name="__main__")
            # use subprocess because run_module does not return control
            subprocess.run(args, check=False)



#! /usr/bin/env python3
"""CoLRev status operation: Display the project status."""
from __future__ import annotations

import io
import typing

import yaml

import colrev.env.utils
import colrev.process.operation
from colrev.constants import Colors
from colrev.constants import OperationsType


class Status(colrev.process.operation.Operation):
    """Determine the status of the project"""

    type = OperationsType.check

    def __init__(self, *, review_manager: colrev.review_manager.ReviewManager) -> None:
        super().__init__(
            review_manager=review_manager,
            operations_type=self.type,
        )

    def get_analytics(self) -> dict:
        """Get status analytics"""

        analytics_dict = {}
        git_repo = self.review_manager.dataset.get_repo()

        revlist = list(
            (
                commit.hexsha,
                commit.message,
                commit.author.name,
                commit.committed_date,
                (
                    commit.tree / str(self.review_manager.paths.STATUS_FILE)
                ).data_stream.read(),
            )
            for commit in git_repo.iter_commits(
                paths=str(self.review_manager.paths.STATUS_FILE)
            )
        )
        for ind, (
            commit_id,
            commit_msg,
            commit_author,
            committed_date,
            filecontents,
        ) in enumerate(revlist):
            var_t = io.StringIO(filecontents.decode("utf-8"))

            # TBD: we could simply include the whole STATUS_FILE
            # (to create a general-purpose status analyzer)
            # -> flatten nested structures (e.g., overall/currently)
            # -> integrate with get_status (current data) -
            # and get_prior? (levels: aggregated_statistics vs. record-level?)

            data_loaded = yaml.safe_load(var_t)
            analytics_dict[len(revlist) - ind] = {
                "atomic_steps": data_loaded["atomic_steps"],
                "completed_atomic_steps": data_loaded["completed_atomic_steps"],
                "commit_id": commit_id,
                "commit_message": commit_msg.split("\n")[0],
                "commit_author": commit_author,
                "committed_date": committed_date,
                "search": data_loaded["overall"]["md_retrieved"],
                "included": data_loaded["overall"]["rev_included"],
            }

        # keys = list(analytics_dict.values())[0].keys()
        # with open("analytics.csv", "w", newline="", encoding="utf8") as output_file:
        #     dict_writer = csv.DictWriter(output_file, keys)
        #     dict_writer.writeheader()
        #     dict_writer.writerows(reversed(analytics_dict.values()))

        return analytics_dict

    def get_review_status_report(
        self, *, records: typing.Optional[dict] = None, colors: bool = True
    ) -> str:
        """Get the review status report"""

        status_stats = self.review_manager.get_status_stats(records=records)

        template = colrev.env.utils.get_template(template_path="ops/commit/status.txt")

        if colors:
            content = template.render(status_stats=status_stats, colors=Colors)
        else:
            content = template.render(status_stats=status_stats, colors=None)

        return content



#! /usr/bin/env python
"""Checker for missing fields."""
from __future__ import annotations

import colrev.env.utils
import colrev.record.qm.quality_model
from colrev.constants import DefectCodes
from colrev.constants import ENTRYTYPES
from colrev.constants import Fields
from colrev.constants import FieldValues

# pylint: disable=too-few-public-methods


class MissingFieldChecker:
    """The MissingFieldChecker"""

    # Based on https://en.wikipedia.org/wiki/BibTeX
    record_field_requirements = {
        ENTRYTYPES.ARTICLE: [
            Fields.AUTHOR,
            Fields.TITLE,
            Fields.JOURNAL,
            Fields.YEAR,
            Fields.VOLUME,
            Fields.NUMBER,
        ],
        ENTRYTYPES.INPROCEEDINGS: [
            Fields.AUTHOR,
            Fields.TITLE,
            Fields.BOOKTITLE,
            Fields.YEAR,
        ],
        ENTRYTYPES.INCOLLECTION: [
            Fields.AUTHOR,
            Fields.TITLE,
            Fields.BOOKTITLE,
            Fields.PUBLISHER,
            Fields.YEAR,
        ],
        ENTRYTYPES.INBOOK: [
            Fields.AUTHOR,
            Fields.TITLE,
            Fields.CHAPTER,
            Fields.PUBLISHER,
            Fields.YEAR,
        ],
        ENTRYTYPES.PROCEEDINGS: [Fields.BOOKTITLE, Fields.EDITOR, Fields.YEAR],
        ENTRYTYPES.CONFERENCE: [Fields.BOOKTITLE, Fields.EDITOR, Fields.YEAR],
        ENTRYTYPES.BOOK: [Fields.AUTHOR, Fields.TITLE, Fields.PUBLISHER, Fields.YEAR],
        ENTRYTYPES.PHDTHESIS: [Fields.AUTHOR, Fields.TITLE, Fields.SCHOOL, Fields.YEAR],
        ENTRYTYPES.BACHELORTHESIS: [
            Fields.AUTHOR,
            Fields.TITLE,
            Fields.SCHOOL,
            Fields.YEAR,
        ],
        ENTRYTYPES.THESIS: [Fields.AUTHOR, Fields.TITLE, Fields.SCHOOL, Fields.YEAR],
        ENTRYTYPES.MASTERSTHESIS: [
            Fields.AUTHOR,
            Fields.TITLE,
            Fields.SCHOOL,
            Fields.YEAR,
        ],
        ENTRYTYPES.TECHREPORT: [
            Fields.AUTHOR,
            Fields.TITLE,
            Fields.INSTITUTION,
            Fields.YEAR,
        ],
        ENTRYTYPES.UNPUBLISHED: [Fields.TITLE, Fields.AUTHOR, Fields.YEAR],
        ENTRYTYPES.MISC: [Fields.AUTHOR, Fields.TITLE, Fields.YEAR],
        ENTRYTYPES.SOFTWARE: [Fields.AUTHOR, Fields.TITLE, Fields.URL],
        ENTRYTYPES.ONLINE: [Fields.AUTHOR, Fields.TITLE, Fields.URL],
    }
    """Fields requirements for respective ENTRYTYPE"""

    # book, inbook: author <- editor

    msg = DefectCodes.MISSING

    def __init__(
        self, quality_model: colrev.record.qm.quality_model.QualityModel
    ) -> None:
        self.quality_model = quality_model

    def run(self, *, record: colrev.record.record.Record) -> None:
        """Run the missing-field checks"""

        if record.data[Fields.ENTRYTYPE] not in self.record_field_requirements:
            return

        if record.masterdata_is_curated():  # pragma: no cover
            return

        required_fields_keys = self.record_field_requirements[
            record.data[Fields.ENTRYTYPE]
        ]
        for required_fields_key in required_fields_keys:
            if record.ignored_defect(key=required_fields_key, defect=self.msg):
                continue

            if self._is_missing(record=record, key=required_fields_key):
                record.update_field(
                    key=required_fields_key,
                    value=FieldValues.UNKNOWN,
                    source="generic_field_requirements",
                )
                record.add_field_provenance_note(key=required_fields_key, note=self.msg)
            else:
                record.remove_field_provenance_note(
                    key=required_fields_key, note=self.msg
                )

    def _is_missing(self, *, key: str, record: colrev.record.record.Record) -> bool:
        if not self._required_in_forthcoming(key=key, record=record):
            return False
        if key in record.data and record.data[key] != FieldValues.UNKNOWN:
            return False
        return True

    def _required_in_forthcoming(
        self, *, key: str, record: colrev.record.record.Record
    ) -> bool:
        if record.data.get(Fields.YEAR, "") != FieldValues.FORTHCOMING:
            return True

        # Forthcoming for the following conditions:
        if key == Fields.YEAR:
            return True

        if key == Fields.VOLUME:
            record.add_field_provenance_note(key=key, note=f"IGNORE:{self.msg}")
            return False
        if key == Fields.NUMBER:
            record.add_field_provenance_note(key=key, note=f"IGNORE:{self.msg}")
            return False

        return True


def register(quality_model: colrev.record.qm.quality_model.QualityModel) -> None:
    """Register the checker"""
    quality_model.register_checker(MissingFieldChecker(quality_model))



#! /usr/bin/env python
"""Convenience functions to load bib files"""
from __future__ import annotations

import io
import itertools
import logging
import os
import re
import string
import typing
from pathlib import Path

import pybtex.errors
from pybtex.database import Person
from pybtex.database.input import bibtex

import colrev.exceptions as colrev_exceptions
import colrev.loader.loader
from colrev.constants import Fields
from colrev.constants import FieldSet
from colrev.constants import FieldValues
from colrev.constants import RecordState


# pylint: disable=too-few-public-methods
# pylint: disable=too-many-arguments


class BIBLoader(colrev.loader.loader.Loader):
    """Loads BibTeX files"""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        *,
        filename: Path,
        unique_id_field: str = "ID",
        entrytype_setter: typing.Callable = lambda x: x,
        field_mapper: typing.Callable = lambda x: x,
        id_labeler: typing.Callable = lambda x: x,
        logger: logging.Logger = logging.getLogger(__name__),
    ):
        super().__init__(
            filename=filename,
            id_labeler=id_labeler,
            unique_id_field=unique_id_field,
            entrytype_setter=entrytype_setter,
            field_mapper=field_mapper,
            logger=logger,
        )

    @classmethod
    def get_nr_records(cls, filename: Path) -> int:
        """Get the number of records in the file"""
        count = 0
        with open(filename, encoding="utf8") as file:
            for line in file:
                if line.startswith("@") and "@comment" not in line[:10].lower():
                    count += 1
        return count

    def _generate_next_unique_id(
        self,
        *,
        temp_id: str,
        existing_ids: list,
    ) -> str:
        """Get the next unique ID"""

        order = 0
        letters = list(string.ascii_lowercase)
        next_unique_id = temp_id
        appends: list = []
        while next_unique_id.lower() in [i.lower() for i in existing_ids]:
            if len(appends) == 0:
                order += 1
                appends = list(itertools.product(letters, repeat=order))
            next_unique_id = temp_id + "".join(list(appends.pop(0)))
        temp_id = next_unique_id
        return temp_id

    def _apply_file_fixes(self) -> None:
        # pylint: disable=too-many-statements

        def fix_key(
            file: typing.IO, line: bytes, replacement_line: bytes, seekpos: int
        ) -> int:
            self.logger.info(
                f"Fix invalid key: \n{line.decode('utf-8')}"
                f"{replacement_line.decode('utf-8')}"
            )
            line = file.readline()
            remaining = line + file.read()
            file.seek(seekpos)
            file.write(replacement_line)
            seekpos = file.tell()
            file.flush()
            os.fsync(file)
            file.write(remaining)
            file.truncate()  # if the replacement is shorter...
            file.seek(seekpos)
            return seekpos

        if self.filename is not None:

            with open(self.filename, encoding="utf8") as bibtex_file:
                contents = bibtex_file.read()
                if len(contents) < 10:
                    return
                bib_r = re.compile(r"@.*{.*,", re.M)
                if len(re.findall(bib_r, contents)) == 0:
                    self.logger.error(f"Not a bib file? {self.filename.name}")
                    raise colrev_exceptions.UnsupportedImportFormatError(self.filename)

            # Errors to fix before pybtex loading:
            # - set_incremental_ids (otherwise, not all records will be loaded)
            # - fix_keys (keys containing white spaces)
            record_ids: typing.List[str] = []
            with open(self.filename, "r+b") as file:
                seekpos = file.tell()
                line = file.readline()
                while line:
                    if b"@" in line[:3]:
                        current_id = line[line.find(b"{") + 1 : line.rfind(b",")]
                        current_id_str = current_id.decode("utf-8").lstrip().rstrip()

                        if any(x in current_id_str for x in [";"]):
                            replacement_line = re.sub(
                                r";",
                                r"_",
                                line.decode("utf-8"),
                            ).encode("utf-8")
                            seekpos = fix_key(file, line, replacement_line, seekpos)

                        if current_id_str in record_ids:
                            next_id = self._generate_next_unique_id(
                                temp_id=current_id_str, existing_ids=record_ids
                            )
                            self.logger.info(
                                f"Fix duplicate ID: {current_id_str} >> {next_id}"
                            )

                            replacement_line = (
                                line.decode("utf-8")
                                .replace(current_id.decode("utf-8"), next_id)
                                .encode("utf-8")
                            )

                            line = file.readline()
                            remaining = line + file.read()
                            file.seek(seekpos)
                            file.write(replacement_line)
                            seekpos = file.tell()
                            file.flush()
                            os.fsync(file)
                            file.write(remaining)
                            file.truncate()  # if the replacement is shorter...
                            file.seek(seekpos)

                            record_ids.append(next_id)

                        else:
                            record_ids.append(current_id_str)

                    # Fix keys
                    if re.match(
                        r"^\s*[a-zA-Z0-9]+\s+[a-zA-Z0-9]+\s*\=", line.decode("utf-8")
                    ):
                        replacement_line = re.sub(
                            r"(^\s*)([a-zA-Z0-9]+)\s+([a-zA-Z0-9]+)(\s*\=)",
                            r"\1\2_\3\4",
                            line.decode("utf-8"),
                        ).encode("utf-8")
                        seekpos = fix_key(file, line, replacement_line, seekpos)

                    # Fix IDs
                    if re.match(
                        r"^@[a-zA-Z0-9]+\{[a-zA-Z0-9]+\s[a-zA-Z0-9]+,",
                        line.decode("utf-8"),
                    ):
                        replacement_line = re.sub(
                            r"^(@[a-zA-Z0-9]+\{[a-zA-Z0-9]+)\s([a-zA-Z0-9]+,)",
                            r"\1_\2",
                            line.decode("utf-8"),
                        ).encode("utf-8")
                        seekpos = fix_key(file, line, replacement_line, seekpos)

                    seekpos = file.tell()
                    line = file.readline()

    def _parse_records_dict(self, *, records_dict: dict) -> dict:
        """Parse a records_dict from pybtex to colrev standard"""

        def format_name(person: Person) -> str:
            def join(name_list: list) -> str:
                return " ".join([name for name in name_list if name])

            first = person.get_part_as_text("first")
            middle = person.get_part_as_text("middle")
            prelast = person.get_part_as_text("prelast")
            last = person.get_part_as_text("last")
            lineage = person.get_part_as_text("lineage")
            name_string = ""
            if last:
                name_string += join([prelast, last])
            if lineage:
                name_string += f", {lineage}"
            if first or middle:
                name_string += ", "
                name_string += join([first, middle])
            return name_string

        # Need to concatenate fields and persons dicts
        # but pybtex is still the most efficient solution.
        records_dict = {
            k: {
                **{Fields.ID: k},
                **{Fields.ENTRYTYPE: v.type},
                **dict(
                    {
                        # Cast status to Enum
                        k: (
                            RecordState[v]
                            if (Fields.STATUS == k)
                            # DOIs are case insensitive -> use upper case.
                            else (
                                v.upper()
                                if (Fields.DOI == k)
                                # Note : the following two lines are a temporary fix
                                # to converg colrev_origins to list items
                                else (
                                    [
                                        el.rstrip().lstrip()
                                        for el in v.split(";")
                                        if "" != el
                                    ]
                                    if k == Fields.ORIGIN
                                    else (
                                        [
                                            el.rstrip()
                                            for el in (v + " ").split("; ")
                                            if "" != el
                                        ]
                                        if k in FieldSet.LIST_FIELDS
                                        else (
                                            self._load_field_dict(value=v, field=k)
                                            if k in [Fields.MD_PROV, Fields.D_PROV]
                                            else v
                                        )
                                    )
                                )
                            )
                        )
                        for k, v in v.fields.items()
                    }
                ),
                **dict(
                    {
                        k: " and ".join(format_name(person) for person in persons)
                        for k, persons in v.persons.items()
                    }
                ),
            }
            for k, v in records_dict.items()
        }

        return records_dict

    def _load_field_dict(self, *, value: str, field: str) -> dict:
        # pylint: disable=too-many-branches

        assert field in [
            Fields.MD_PROV,
            Fields.D_PROV,
        ], f"error loading dict_field: {field}"

        return_dict: typing.Dict[str, typing.Any] = {}
        if value == "":  # pragma: no cover
            return return_dict

        if field == Fields.MD_PROV:
            if value[:7] == FieldValues.CURATED:
                source = value[value.find(":") + 1 : value[:-1].rfind(";")]
                return_dict[FieldValues.CURATED] = {
                    "source": source,
                    "note": "",
                }

            else:
                # Pybtex automatically replaces \n in fields.
                # For consistency, we also do that for header_only mode:
                if "\n" in value:  # pragma: no cover
                    value = value.replace("\n", " ")
                items = [x.lstrip() + ";" for x in (value + " ").split("; ") if x != ""]

                for item in items:
                    key_source = item[: item[:-1].rfind(";")]
                    assert (
                        ":" in key_source
                    ), f"problem with masterdata_provenance_item {item}"
                    note = item[item[:-1].rfind(";") + 1 : -1]
                    key, source = key_source.split(":", 1)
                    # key = key.rstrip().lstrip()
                    return_dict[key] = {
                        "source": source,
                        "note": note,
                    }

        elif field == Fields.D_PROV:
            # Note : pybtex replaces \n upon load
            for item in (value + " ").split("; "):
                if item == "":
                    continue
                item += ";"  # removed by split
                key_source = item[: item[:-1].rfind(";")]
                note = item[item[:-1].rfind(";") + 1 : -1]
                assert ":" in key_source, f"problem with data_provenance_item {item}"
                key, source = key_source.split(":", 1)
                return_dict[key] = {
                    "source": source,
                    "note": note,
                }

        return return_dict

    def _parse_k_v(self, item_string: str) -> tuple:
        if " = " in item_string:
            key, value = item_string.split(" = ", 1)
        else:
            key = Fields.ID
            value = item_string.split("{")[1]

        key = key.lstrip().rstrip()
        value = value.lstrip().rstrip().lstrip("{").rstrip("},")
        if key == Fields.ORIGIN:
            value_list = value.replace("\n", "").split(";")
            value_list = [x.lstrip(" ").rstrip(" ") for x in value_list if x]
            return key, value_list
        if key == Fields.STATUS:
            return key, RecordState[value]
        if key == Fields.MD_PROV:
            return key, self._load_field_dict(value=value, field=key)
        if key == Fields.FILE:
            return key, Path(value)

        return key, value

    # pylint: disable=too-many-branches
    def _read_record_header_items(
        self, *, file_object: typing.Optional[typing.TextIO] = None
    ) -> list:
        # Note : more than 10x faster than the pybtex part of load_records_dict()

        if file_object is None:
            assert self.filename is not None
            # pylint: disable=consider-using-with
            file_object = open(self.filename, encoding="utf-8")

        # Fields required
        default = {
            Fields.ID: "NA",
            Fields.ORIGIN: "NA",
            Fields.STATUS: "NA",
            Fields.FILE: "NA",
            Fields.SCREENING_CRITERIA: "NA",
            Fields.MD_PROV: "NA",
        }
        # number_required_header_items = len(default)

        record_header_item = default.copy()
        item_count, item_string, record_header_items = (
            0,
            "",
            [],
        )
        while True:
            line = file_object.readline()
            if not line:
                break

            if line[:1] == "%" or line == "\n":
                continue

            # if item_count > number_required_header_items or "}" == line:
            #     record_header_items.append(record_header_item)
            #     record_header_item = default.copy()
            #     item_count = 0
            #     continue

            if "@" in line[:2] and record_header_item[Fields.ID] != "NA":
                record_header_items.append(record_header_item)
                record_header_item = default.copy()
                item_count = 0

            item_string += line
            if "}," in line or "@" in line[:2]:
                key, value = self._parse_k_v(item_string)
                # if key == Fields.MD_PROV:
                #     if value == "NA":
                #         value = {}
                # if value == "NA":
                #     item_string = ""
                #     continue
                item_string = ""
                if key in record_header_item:
                    item_count += 1
                    record_header_item[key] = value

        if record_header_item[Fields.ORIGIN] != "NA":
            record_header_items.append(record_header_item)

        return [
            {k: v for k, v in record_header_item.items() if "NA" != v}
            for record_header_item in record_header_items
        ]

    def get_record_header_items(self) -> dict:
        """Get the record header items"""
        record_header_list = []
        record_header_list = self._read_record_header_items()

        record_header_dict = {r[Fields.ID]: r for r in record_header_list}
        return record_header_dict

    def load_records_list(self) -> list:

        def drop_empty_fields(*, records: dict) -> None:
            for record_id in records:
                records[record_id] = {
                    k: v
                    for k, v in records[record_id].items()
                    if v is not None and v != "nan"
                }

        def resolve_crossref(*, records: dict) -> None:
            # https://bibtex.eu/fields/crossref/
            crossref_ids = []
            for record_dict in records.values():
                if "crossref" not in record_dict:
                    continue

                crossref_record = records.get(record_dict["crossref"], None)

                if not crossref_record:
                    self.logger.error(
                        f"crossref record (ID={record_dict['crossref']}) not found"
                    )
                    continue
                crossref_ids.append(crossref_record["ID"])
                for key, value in crossref_record.items():
                    if key not in record_dict:
                        record_dict[key] = value
                del record_dict["crossref"]

            for crossref_id in crossref_ids:
                del records[crossref_id]

        self._apply_file_fixes()

        temp_f = io.StringIO()
        pybtex.io.stderr = temp_f
        pybtex.errors.set_strict_mode(False)
        parser = bibtex.Parser()
        bib_data = parser.parse_file(str(self.filename))
        records = self._parse_records_dict(records_dict=bib_data.entries)

        drop_empty_fields(records=records)
        resolve_crossref(records=records)
        records = dict(sorted(records.items()))

        return list(records.values())



#! /usr/bin/env python
"""SearchSource: Semantic Scholar"""
from __future__ import annotations

import typing
from multiprocessing import Lock
from pathlib import Path

import requests
import zope.interface
from pydantic import Field
from semanticscholar import SemanticScholar
from semanticscholar import SemanticScholarException
from semanticscholar.PaginatedResults import PaginatedResults

import colrev.constants
import colrev.env.environment_manager
import colrev.env.language_service
import colrev.exceptions as colrev_exceptions
import colrev.ops.load
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.process.operation
import colrev.record.record
import colrev.settings
from colrev.constants import Colors
from colrev.constants import Fields
from colrev.constants import SearchSourceHeuristicStatus
from colrev.constants import SearchType
from colrev.packages.semanticscholar.src import record_transformer
from colrev.packages.semanticscholar.src.semanticscholar_ui import SemanticScholarUI

if typing.TYPE_CHECKING:  # pragma: no cover

    import colrev.ops.prep


# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.package_manager.interfaces.SearchSourceInterface)
class SemanticScholarSearchSource:
    """Semantic Scholar API Search Source"""

    settings_class = colrev.package_manager.package_settings.DefaultSourceSettings
    # Provide objects with classes
    _s2: SemanticScholar
    _search_return: PaginatedResults

    endpoint = "colrev.semanticscholar"
    ci_supported: bool = Field(default=True)

    # SearchSourceInterface constants
    heuristic_status = SearchSourceHeuristicStatus.oni
    search_types = [SearchType.API]

    source_identifier = Fields.SEMANTIC_SCHOLAR_ID

    SETTINGS = {
        "api_key": "packages.search_source.colrev.semanticscholar.api_key",
    }

    _availability_exception_message = (
        f"Semantic Scholar ({Colors.ORANGE}check "
        f"https://status.api.semanticscholar.org/{Colors.END})"
    )

    _s2_UI = SemanticScholarUI()
    _s2_filename = Path("data/search/md_semscholar.bib")

    def __init__(
        self,
        *,
        source_operation: colrev.process.operation.Operation,
        settings: typing.Optional[dict] = None,
    ) -> None:
        self.review_manager = source_operation.review_manager
        if settings:
            # Semantic Scholar as a search source
            self.search_source = self.settings_class(**settings)
        else:
            self.search_source = colrev.settings.SearchSource(
                endpoint="colrev.semanticscholar",
                filename=self._s2_filename,
                search_type=SearchType.API,
                search_parameters={},
                comment="",
            )
            self.s2_lock = Lock()

    def check_availability(
        self, *, source_operation: colrev.process.operation.Operation
    ) -> None:
        """Check the availability of the Semantic Scholar API"""

        try:
            # pylint: disable=duplicate-code
            test_doi = "10.17705/1CAIS.04607"
            test_record = {
                "title": "A Knowledge Development Perspective on Literature Reviews: "
                + "Validation of a new Typology in the IS Field",
                "url": "https://www.semanticscholar.org/paper"
                "/b639a05d936dfd519fe4098edc95b5680b7ec7ec",
            }

            returned_record = self._s2.get_paper(paper_id=test_doi)

            if 0 != len(returned_record):
                assert returned_record[Fields.TITLE] == test_record[Fields.TITLE]
                assert returned_record[Fields.URL] == test_record[Fields.URL]
            else:
                if not self.review_manager.force_mode:
                    raise colrev_exceptions.ServiceNotAvailableException(
                        self._availability_exception_message
                    )
        except (requests.exceptions.RequestException, IndexError) as exc:
            print(exc)
            if not self.review_manager.force_mode:
                raise colrev_exceptions.ServiceNotAvailableException(
                    self._availability_exception_message
                ) from exc

    def _get_semantic_scholar_api(
        self, *, params: dict, rerun: bool
    ) -> PaginatedResults:
        """Get Semantic Scholar API depending on
        the search subject and look for search parameters"""
        subject = params.pop("search_subject")
        if subject == "keyword":
            _search_return = self.keyword_search(params=params, rerun=rerun)
        elif subject == "paper":
            _search_return = self.paper_search(params=params, rerun=rerun)
        elif subject == "author":
            _search_return = self.author_search(params=params, rerun=rerun)
        else:
            self.review_manager.logger.info("No search parameters were found.")
            raise NotImplementedError

        return _search_return

    def keyword_search(self, *, params: dict, rerun: bool) -> PaginatedResults:
        """Prepare search parameters and conduct full keyword search with the python client"""

        query = None
        year = None
        venue = None
        fields_of_study = None
        open_access_pdf = None

        for key, value in params.items():
            if key == "query":
                query = value
            elif key == "year":
                year = value
            elif key == "venue":
                venue = value
            elif key == "fields_of_study":
                fields_of_study = value
            elif key == "open_access_pdf":
                open_access_pdf = value

        try:
            record_return = self._s2.search_paper(
                query=query,
                year=year,
                publication_types=None,
                venue=venue,
                fields_of_study=fields_of_study,
                open_access_pdf=open_access_pdf,
            )
        except (
            SemanticScholarException.SemanticScholarException,
            SemanticScholarException.BadQueryParametersException,
        ) as exc:
            self.review_manager.logger.error(
                "Error: Something went wrong during the search with the Python Client."
                + " This program will close."
            )
            print(exc)
            raise colrev_exceptions.ServiceNotAvailableException(exc)

        self.review_manager.logger.info(
            str(record_return.total) + " records have been found.\n"
        )

        if record_return.total == 0:
            self.review_manager.logger.info(
                "Search aborted because no records were found. This program will close."
            )
            raise colrev_exceptions.ServiceNotAvailableException(
                "No records were found."
            )

        return record_return

    def paper_search(self, *, params: dict, rerun: bool) -> PaginatedResults:
        """Method to conduct an API search of SemanticScholar with a List of Paper IDs"""

        if "paper_ids" in params:
            try:
                record_return = self._s2.get_papers(params.get("paper_ids"))
            except (
                SemanticScholarException.SemanticScholarException,
                SemanticScholarException.BadQueryParametersException,
            ) as exc:
                self.review_manager.logger.error(
                    "Error: Something went wrong during the search with the Python Client."
                    + " This program will close."
                )
                print(exc)
                raise colrev_exceptions.ServiceNotAvailableException(exc)
        else:
            self.review_manager.logger.error(
                'Error: Search type "Search for paper" is not available with your parameters.\n'
                + "Search parameter: "
                + str(params.get("paper_ids"))
            )
            raise colrev_exceptions.ServiceNotAvailableException(
                "Search parameter error."
            )
        return record_return

    def author_search(self, *, params: dict, rerun: bool) -> PaginatedResults:
        """Method to conduct an API search of SemanticScholar with a List of Author IDs"""

        if "author_ids" in params:
            try:
                record_return = self._s2.get_authors(params.get("author_ids"))
            except (
                SemanticScholarException.SemanticScholarException,
                SemanticScholarException.BadQueryParametersException,
            ) as exc:
                self.review_manager.logger.error(
                    "Error: Something went wrong during the search with the Python Client."
                    + " This program will close."
                )
                print(exc)
                raise colrev_exceptions.ServiceNotAvailableException(exc)
        else:
            self.review_manager.logger.error(
                '\nError: Search type "Search for author" is not available with your parameters.\n'
                + "Search parameter: "
                + str(params.get("author_ids"))
            )
            raise colrev_exceptions.ServiceNotAvailableException(
                "Search parameter error."
            )
        return record_return

    def _get_api_key(self) -> str:
        """Method to request an API key from the settings file - or, if empty, from user input"""
        api_key = self.review_manager.environment_manager.get_settings_by_key(
            self.SETTINGS["api_key"]
        )

        if api_key:
            api_key = self._s2_UI.get_api_key(api_key)
        else:
            api_key = self._s2_UI.get_api_key()

        if api_key:
            self.review_manager.environment_manager.update_registry(
                self.SETTINGS["api_key"], api_key
            )
        else:
            self.review_manager.environment_manager.update_registry(
                self.SETTINGS["api_key"], ""
            )

        return api_key

    def search(self, rerun: bool) -> None:
        """Run a search of Semantic Scholar"""

        # get the api key
        s2_api_key = self._get_api_key()
        if s2_api_key:
            self._s2 = SemanticScholar(api_key=s2_api_key)
        else:
            self._s2 = SemanticScholar()

        # load file because the object is not shared between processes
        s2_feed = self.search_source.get_api_feed(
            review_manager=self.review_manager,
            source_identifier=self.source_identifier,
            update_only=(not rerun),
        )
        # rerun not implemented yet
        if rerun:
            self.review_manager.logger.info(
                "Performing a search of the full history (may take time)"
            )
        try:
            params = self.search_source.search_parameters
            _search_return = self._get_semantic_scholar_api(params=params, rerun=rerun)

        except SemanticScholarException.BadQueryParametersException as exc:
            self.review_manager.logger.error(
                "Error: Invalid Search Parameters. The Program will close."
            )
            print(exc)
            raise colrev_exceptions.ServiceNotAvailableException(exc)
            # KeyError  # error in semanticscholar package:
        except (KeyError, PermissionError) as exc:
            self.review_manager.logger.error(
                "Error: Search could not be conducted. Please check your Parameters and API key."
            )
            print(exc)
            raise colrev_exceptions.ServiceNotAvailableException("No valid API key.")

        try:
            for item in _search_return:
                retrieved_record = record_transformer.dict_to_record(item=item)
                s2_feed.add_update_record(retrieved_record)

        except (
            colrev_exceptions.RecordNotParsableException,
            colrev_exceptions.NotFeedIdentifiableException,
        ) as exc:
            self.review_manager.logger.error(
                "Error: Retrieved records were not parsable into feed and result file."
            )
            print(exc)

        s2_feed.save()

    @classmethod
    def add_endpoint(
        cls,
        operation: colrev.ops.search.Search,
        params: str,
    ) -> colrev.settings.SearchSource:
        """Add SearchSource as an endpoint (based on query provided to colrev search --add )"""

        # get search parameters from the user interface
        cls._s2_UI.main_ui()
        search_subject = cls._s2_UI.search_subject
        search_params = cls._s2_UI.search_params

        search_params["search_subject"] = search_subject

        # decrease probability to reach linux file name length limit
        # by only using three parameters for result file
        short_search_params_key_list = [
            "query",
            "paper_ids",
            "author_ids",
            "year",
            "open_access_pdf",
        ]
        short_search_params = {}

        if search_params:
            for key, value in search_params.items():
                if key in short_search_params_key_list:
                    short_search_params[key] = value

        if "query" in short_search_params:
            if len(short_search_params["query"]) > 50:
                short_search_params["query"] = short_search_params["query"][0:50]

        filename = operation.get_unique_filename(file_path_string="semanticscholar")

        search_source = colrev.settings.SearchSource(
            endpoint="colrev.semanticscholar",
            filename=filename,
            search_type=SearchType.API,
            search_parameters=search_params,
            comment="",
        )
        operation.add_source_and_search(search_source)
        return search_source

    def prep_link_md(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.record.Record,
        save_feed: bool = True,
        timeout: int = 30,
    ) -> colrev.record.record.Record:
        """Retrieve master data from Semantic Scholar"""
        # Not yet implemented
        return record

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for Semantic Scholar"""
        # Not yet implemented

        result = {"confidence": 0.0}

        return result

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""
        # Not yet implemented

        if self.search_source.filename.suffix == ".bib":
            records = colrev.loader.load_utils.load(
                filename=self.search_source.filename,
                logger=self.review_manager.logger,
            )
            return records

        raise NotImplementedError

    def prepare(
        self,
        record: colrev.record.record_prep.PrepRecord,
        source: colrev.settings.SearchSource,
    ) -> colrev.record.record_prep.PrepRecord:
        """Source-specific preparation for Semantic Scholar"""
        # Not yet implemented
        return record



#! /usr/bin/env python
"""ERIC API"""
import typing

import requests

import colrev.env.language_service
import colrev.exceptions as colrev_exceptions
import colrev.record.record_prep
from colrev.constants import ENTRYTYPES
from colrev.constants import Fields


# pylint: disable=too-few-public-methods


class ERICAPI:
    """Connector for the ERIC API"""

    API_FIELDS = [
        "title",
        "author",
        "source",
        "publicationdateyear",
        "description",
        "subject",
        "peerreviewed",
        "abstractor",
        "audience",
        "authorxlink",
        "e_datemodified",
        "e_fulltextauth",
        "e yearadded",
        "educationlevel",
        "identifiersgeo",
        "identifierslaw",
        "identifierstest",
        "iescited",
        "iesfunded",
        "iesgrantcontractnum",
        "iesgrantcontractnumxlink",
        "ieslinkpublication",
        "ieslinkwwcreviewguide",
        "ieswwcreviewed",
        "institution",
        "isbn",
        "issn",
        "language",
        "publisher",
        "sourceid",
        "sponsor",
        "url",
    ]
    FIELD_MAPPING = {"publicationdateyear": "year", "description": "abstract"}

    def __init__(self, params: dict) -> None:
        self.params = params
        self.language_service = colrev.env.language_service.LanguageService()

    def _build_search_url(self) -> str:
        url = "https://api.ies.ed.gov/eric/"
        query = self.params["query"]
        format_param = "json"
        start_param = self.params.get("start", "0")
        rows_param = self.params.get("rows", "2000")
        return f"{url}?search={query}&format={format_param}&start={start_param}&rows={rows_param}"

    def _create_record(self, doc: dict) -> colrev.record.record.Record:
        # pylint: disable=too-many-branches
        record_dict = {Fields.ID: doc["id"]}
        record_dict[Fields.ENTRYTYPE] = "other"
        if "Journal Articles" in doc["publicationtype"]:
            record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.ARTICLE
        elif "Books" in doc["publicationtype"]:
            record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.BOOK

        for field in self.API_FIELDS:
            field_value = doc.get(field)
            if field_value is not None:
                record_dict[field] = field_value

        for api_field, rec_field in self.FIELD_MAPPING.items():
            if api_field not in record_dict:
                continue
            record_dict[rec_field] = record_dict.pop(api_field)

        if "source" in doc:
            record_dict[Fields.JOURNAL] = doc.pop("source")

        if "subject" in record_dict:
            record_dict["subject"] = ", ".join(record_dict["subject"])

        if Fields.AUTHOR in record_dict:
            # pylint: disable=colrev-missed-constant-usage
            record_dict[Fields.AUTHOR] = " and ".join(record_dict["author"])
        if Fields.ISSN in record_dict:
            # pylint: disable=colrev-missed-constant-usage
            record_dict[Fields.ISSN] = record_dict["issn"][0].lstrip("EISSN-")
        if Fields.ISBN in record_dict:
            # pylint: disable=colrev-missed-constant-usage
            record_dict[Fields.ISBN] = record_dict["isbn"][0].lstrip("ISBN-")

        if Fields.YEAR in record_dict:
            # pylint: disable=colrev-missed-constant-usage
            record_dict[Fields.YEAR] = str(record_dict["year"])

        record = colrev.record.record.Record(record_dict)
        if Fields.LANGUAGE in record.data:
            try:
                record.data[Fields.LANGUAGE] = record.data[Fields.LANGUAGE][0]
                self.language_service.unify_to_iso_639_3_language_codes(record=record)
            except colrev_exceptions.InvalidLanguageCodeException:
                del record.data[Fields.LANGUAGE]
        return record

    def get_query_return(self) -> typing.Iterator[colrev.record.record.Record]:
        """Get the records from a query"""
        full_url = self._build_search_url()

        response = requests.get(full_url, timeout=90)
        if response.status_code != 200:
            return
        with open("test.json", "wb") as file:
            file.write(response.content)
        data = response.json()

        if "docs" not in data["response"]:
            raise colrev_exceptions.ServiceNotAvailableException(
                "Could not reach API. Status Code: " + response.status_code
            )

        for doc in data["response"]["docs"]:
            record = self._create_record(doc)
            yield record
